#!/usr/bin/env python

# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Program Name>
  in_toto_run.py

<Author>
  Lukas Puehringer <lukas.puehringer@nyu.edu>

<Started>
  June 27, 2016

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides a command line interface for runlib.in_toto_run.

<Return Codes>
  2 if an exception occurred during argument parsing
  1 if an exception occurred
  0 if no exception occurred

"""

import sys
import argparse
import logging
import in_toto.user_settings
from in_toto import (runlib, __version__)

from in_toto.common_args import (EXCLUDE_ARGS, EXCLUDE_KWARGS, BASE_PATH_ARGS,
    BASE_PATH_KWARGS, LSTRIP_PATHS_ARGS, LSTRIP_PATHS_KWARGS, KEY_ARGS,
    KEY_KWARGS, KEY_TYPE_KWARGS, KEY_TYPE_ARGS, GPG_ARGS, GPG_KWARGS,
    GPG_HOME_ARGS, GPG_HOME_KWARGS, VERBOSE_ARGS, VERBOSE_KWARGS, QUIET_ARGS,
    QUIET_KWARGS, METADATA_DIRECTORY_ARGS, METADATA_DIRECTORY_KWARGS,
    KEY_PASSWORD_ARGS, KEY_PASSWORD_KWARGS, parse_password_and_prompt_args,
    sort_action_groups, title_case_action_groups, OPTS_TITLE)

from securesystemslib import interface

# Command line interfaces should use in_toto base logger (c.f. in_toto.log)
LOG = logging.getLogger("in_toto")


def create_parser():
  """Create and return configured ArgumentParser instance. """
  parser = argparse.ArgumentParser(
      formatter_class=argparse.RawDescriptionHelpFormatter,
      description="""
in-toto-run is the main command line interface for generating link
metadata while carrying out a supply chain step. To do this, it wraps the
passed command, and attempts to track all relevant information about the
wrapped command's execution. It records paths and hashes of 'materials' (files
before command execution) and 'products' (files after command execution) and
writes them together with other information (executed command, return value,
stdout and stderr) to a link metadata file, which is signed with the passed
key. It returns a non-zero value on failure and zero otherwise.""")

  parser.usage = (
    "%(prog)s <named arguments> [{}] \\\n\t -- <command> [args]".format(
      OPTS_TITLE.lower()))

  parser.epilog = """EXAMPLE USAGE

Tag a git repo, storing files in CWD as products, signing the resulting link
file with the private key loaded from 'key_file'.

  {prog} -n tag -p . -k key_file -- git tag v1.0


Create a tarball, storing files in 'project' directory as materials and the
tarball as product, signing the link file with a GPG key '...7E0C8A17'.

  {prog} -n package -m project -p project.tar.gz \\
         -g 8465A1E2E0FB2B40ADB2478E18FB3F537E0C8A17 \\
         -- tar czf project.tar.gz project


Not all supply chain steps require that a command be executed. in-toto can
still generate signed attestations, e.g. for review work. In that case, files
may be marked as materials for the manual review process and the command be
omitted.

  {prog} -n review -k key_file -m document.pdf -x


If an artifact that should be recorded is not in the current working directory
(or one of its subdirectories) it can be located using the base path option.
Note that in this example only the relative path, 'document.pdf' is stored
along with its hash in the resulting link metadata file.

  {prog} -n review -k key_file -m document.pdf \\
         --base-path /my/review/docs/ -x


Similarly, it is possible to pass the full path to the artifact that should
be recorded together with a left-strip path, to only store a relative path,
e.g. 'document.pdf'.

  {prog} -n review -k key_file \\
         -m /tmp/my/review/docs/document.pdf \\
         --lstrip-paths /tmp/my/review/docs/ -x


""".format(prog=parser.prog)


  named_args = parser.add_argument_group("required named arguments")

  # FIXME: Do we limit the allowed characters for the name?
  named_args.add_argument("-n", "--step-name", type=str, required=True,
      metavar="<name>", help=(
      "name for the resulting link metadata file, which is written to"
      " '<name>.<keyid prefix>.link'. It is also used to associate the link"
      " with a step defined in an in-toto layout."))

  parser.add_argument("-m", "--materials", type=str, required=False,
      nargs='+', metavar="<path>", help=(
      "paths to files or directories, for which paths and hashes are stored in"
      " the resulting link metadata before the command is executed. Symlinks"
      " to files are followed."))

  parser.add_argument("-p", "--products", type=str, required=False,
      nargs='+', metavar="<path>", help=(
      "paths to files or directories, for which paths and hashes are stored in"
      " the resulting link metadata after the command is executed. Symlinks to"
      " files are followed."))

  parser.add_argument("-s", "--record-streams", dest="record_streams",
      default=False, action="store_true", help=(
      "duplicate 'stdout' and 'stderr' of the executed command and store the"
      " contents in the resulting link metadata. Do not use with interactive"
      " commands."))

  parser.add_argument("-x", "--no-command", dest="no_command", default=False,
    action="store_true", help=(
    "generate link metadata without executing a command, e.g. for a"
    " signed-off-by step."))

  named_args.add_argument(*KEY_ARGS, **KEY_KWARGS)
  parser.add_argument(*KEY_TYPE_ARGS, **KEY_TYPE_KWARGS)
  parser.add_argument(*KEY_PASSWORD_ARGS, **KEY_PASSWORD_KWARGS)

  named_args.add_argument(*GPG_ARGS, **GPG_KWARGS)
  parser.add_argument(*GPG_HOME_ARGS, **GPG_HOME_KWARGS)

  parser.add_argument(*EXCLUDE_ARGS, **EXCLUDE_KWARGS)
  parser.add_argument(*BASE_PATH_ARGS, **BASE_PATH_KWARGS)
  parser.add_argument(*LSTRIP_PATHS_ARGS, **LSTRIP_PATHS_KWARGS)
  parser.add_argument(*METADATA_DIRECTORY_ARGS, **METADATA_DIRECTORY_KWARGS)

  verbosity_args = parser.add_mutually_exclusive_group(required=False)
  verbosity_args.add_argument(*VERBOSE_ARGS, **VERBOSE_KWARGS)
  verbosity_args.add_argument(*QUIET_ARGS, **QUIET_KWARGS)


  # FIXME: This is not yet ideal.
  # What should we do with tokens like > or ; ?
  parser.add_argument("link_cmd", nargs="*", metavar="<command>",
      help=(
      "command to be executed. It is separated from named and optional"
      " arguments by a double dash '--'."))

  parser.add_argument('--version', action='version',
                      version='{} {}'.format(parser.prog, __version__))

  title_case_action_groups(parser)
  sort_action_groups(parser)

  return parser


def main():
  """Parse arguments, load key from disk (prompts for password if key is
  encrypted) and call in_toto_run.  """
  parser = create_parser()
  args = parser.parse_args()

  LOG.setLevelVerboseOrQuiet(args.verbose, args.quiet)

  # Override defaults in settings.py with environment variables and RCfiles
  in_toto.user_settings.set_settings()

  # Regular signing and GPG signing are mutually exclusive
  if (args.key is None) == (args.gpg is None):
    parser.print_usage()
    parser.error("Specify either '--key <key path>' or '--gpg [<keyid>]'")

  password, prompt = parse_password_and_prompt_args(args)

  # If `--gpg` was set without argument it has the value `True` and
  # we will try to sign with the default key
  gpg_use_default = (args.gpg is True)

  # Otherwise we interpret it as actual keyid
  gpg_keyid = None
  if args.gpg is not True:
    gpg_keyid = args.gpg

  # If no_command is specified run in_toto_run without executing a command
  if args.no_command:
    args.link_cmd = []

  elif not args.link_cmd: # pragma: no branch
    parser.print_usage()
    parser.error("No command specified."
        " Please specify (or use the --no-command option)")

  try:
    # We load the key here because it might prompt the user for a password in
    # case the key is encrypted. Something that should not happen in the lib.
    key = None
    if args.key:
      key = interface.import_privatekey_from_file(
          args.key, key_type=args.key_type, password=password, prompt=prompt)

    runlib.in_toto_run(
        args.step_name, args.materials, args.products, args.link_cmd,
        record_streams=args.record_streams, signing_key=key,
        gpg_keyid=gpg_keyid, gpg_use_default=gpg_use_default,
        gpg_home=args.gpg_home, exclude_patterns=args.exclude_patterns,
        base_path=args.base_path, lstrip_paths=args.lstrip_paths,
        metadata_directory=args.metadata_directory)

  except Exception as e:
    LOG.error("(in-toto-run) {0}: {1}".format(type(e).__name__, e))
    sys.exit(1)

  sys.exit(0)


if __name__ == "__main__":
  main()



# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
Configure base logger for in_toto (see in_toto.log for details).

"""
import in_toto.log
from securesystemslib import KEY_TYPE_RSA, KEY_TYPE_ED25519

SUPPORTED_KEY_TYPES = [KEY_TYPE_RSA, KEY_TYPE_ED25519]


# in-toto version
__version__ = "1.2.0"



#!/usr/bin/env python

# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Program Name>
  link.py

<Author>
  Lukas Puehringer <lukas.puehringer@nyu.edu>
  Santiago Torres <santiago@nyu.edu>

<Started>
  Sep 23, 2016

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides a class for link metadata which is information gathered when a
  step of the supply chain is performed.
"""

import attr
import securesystemslib.formats
from in_toto.models.common import Signable


FILENAME_FORMAT = "{step_name}.{keyid:.8}.link"
FILENAME_FORMAT_SHORT = "{step_name}.link"
UNFINISHED_FILENAME_FORMAT = ".{step_name}.{keyid:.8}.link-unfinished"
UNFINISHED_FILENAME_FORMAT_GLOB = ".{step_name}.{pattern}.link-unfinished"


@attr.s(repr=False, init=False)
class Link(Signable):
  """Evidence for a performed step or inspection of the supply chain.

  A Link object is usually contained in a generic Metablock object for signing,
  serialization and I/O capabilities.

  Attributes:
    name: A unique name used to identify the related step or inspection in an
        in-toto layout.

    command: A list of command and command arguments that report how the
        corresponding step is performed.

    materials: A dictionary of the artifacts *used* by the step, i.e::

            {
              "<material path>": {
                "<hash algorithm name>": "<hash digest of material>",
                ...
              },
              ...
            }

    products: A dictionary of the artifacts *produced* by the step, i.e::

            {
              "<product path>": {
                "<hash algorithm name>": "<hash digest of product>",
                ...
              },
              ...
            }

    byproducts: An opaque dictionary that lists byproducts of the link command
        execution. It should have at least the following entries
        "stdout" (str), "stderr" (str) and "return-value" (int).

    environment: An opaque dictionary that lists information about the
        execution environment of the link command. eg.::

            {
              "variables": "<list of env var KEY=value pairs>",
              "filesystem": "<filesystem info>",
              "workdir": "<CWD when executing link command>"
            }

  """
  _type = attr.ib()
  name = attr.ib()
  materials = attr.ib()
  products = attr.ib()
  byproducts = attr.ib()
  command = attr.ib()
  environment = attr.ib()


  def __init__(self, **kwargs):
    super(Link, self).__init__()

    self._type = "link"
    self.name = kwargs.get("name")
    self.materials = kwargs.get("materials", {})
    self.products = kwargs.get("products", {})
    self.byproducts = kwargs.get("byproducts", {})
    self.command = kwargs.get("command", [])
    self.environment = kwargs.get("environment", {})

    self.validate()

  @property
  def type_(self):
    """The string "link" to indentify the in-toto metadata type."""
    # NOTE: We expose the type_ property in the API documentation instead of
    # _type to protect it against modification.
    # NOTE: Trailing underscore is used by convention (pep8) to avoid conflict
    # with Python's type keyword.
    return self._type

  @staticmethod
  def read(data):
    """Creates a Link object from its dictionary representation.

    Arguments:
      data: A dictionary with link metadata fields.

    Raises:
      securesystemslib.exceptions.FormatError: Passed data is invalid.

    Returns:
      The created Link object.

    """
    return Link(**data)


  def _validate_type(self):
    """Private method to check that `_type` is set to "link"."""
    if self._type != "link":
      raise securesystemslib.exceptions.FormatError(
          "Invalid Link: field `_type` must be set to 'link', got: {}"
          .format(self._type))


  def _validate_materials(self):
    """Private method to check that `materials` is a `dict` of `HASHDICTs`."""
    if not isinstance(self.materials, dict):
      raise securesystemslib.exceptions.FormatError(
          "Invalid Link: field `materials` must be of type dict, got: {}"
          .format(type(self.materials)))

    for material in list(self.materials.values()):
      securesystemslib.formats.HASHDICT_SCHEMA.check_match(material)


  def _validate_products(self):
    """Private method to check that `products` is a `dict` of `HASHDICTs`."""
    if not isinstance(self.products, dict):
      raise securesystemslib.exceptions.FormatError(
          "Invalid Link: field `products` must be of type dict, got: {}"
          .format(type(self.products)))

    for product in list(self.products.values()):
      securesystemslib.formats.HASHDICT_SCHEMA.check_match(product)


  def _validate_byproducts(self):
    """Private method to check that `byproducts` is a `dict`."""
    if not isinstance(self.byproducts, dict):
      raise securesystemslib.exceptions.FormatError(
          "Invalid Link: field `byproducts` must be of type dict, got: {}"
          .format(type(self.byproducts)))


  def _validate_command(self):
    """Private method to check that `command` is a `list`."""
    if not isinstance(self.command, list):
      raise securesystemslib.exceptions.FormatError(
          "Invalid Link: field `command` must be of type list, got: {}"
          .format(type(self.command)))


  def _validate_environment(self):
    """Private method to check that `environment` is a `dict`. """
    if not isinstance(self.environment, dict):
      raise securesystemslib.exceptions.FormatError(
          "Invalid Link: field `environment` must be of type dict, got: {}"
          .format(type(self.environment)))



# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Program Name>
  user_settings.py

<Author>
  Lukas Puehringer <lukas.puehringer@nyu.edu>

<Started>
  Oct 25, 2017

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides methods to parse environment variables (`get_env`) and RCfiles
  (`get_rc`) and to override default settings (`set_settings`) defined in the
  `in_toto.settings` module.

  Check out the respective docstrings to learn about the requirements for
  environment variables and RCfiles (includes examples).

"""
import os
import logging
import in_toto.settings

try:
  import configparser
except ImportError: # pragma: no cover
  import ConfigParser as configparser

# Inherits from in_toto base logger (c.f. in_toto.log)
LOG = logging.getLogger(__name__)


USER_PATH = os.path.expanduser("~")

# Prefix required by environment variables to be considered as in_toto settings
ENV_PREFIX = "IN_TOTO_"

# List of considered rcfile paths in the order they get parsed and overridden,
# i.e. the same setting in `/etc/in_toto/config` and `.in_totorc` (cwd) uses
# the latter
RC_PATHS = [
  os.path.join("/etc", "in_toto", "config"),
  os.path.join("/etc", "in_totorc"),
  os.path.join(USER_PATH, ".config", "in_toto", "config"),
  os.path.join(USER_PATH, ".config", "in_toto"),
  os.path.join(USER_PATH, ".in_toto", "config"),
  os.path.join(USER_PATH, ".in_totorc"),
  ".in_totorc"
]

# List of settings, for which defaults exist in `settings.py`
# TODO: Should we use `dir` on the module instead? If we list them here, we
# have to manually update if `settings.py` changes.
IN_TOTO_SETTINGS = [
  "ARTIFACT_EXCLUDE_PATTERNS", "ARTIFACT_BASE_PATH", "LINK_CMD_EXEC_TIMEOUT"
]


def _colon_split(value):
  """ If `value` contains colons, return a list split at colons,
  return value otherwise. """
  value_list = value.split(":")
  if len(value_list) > 1:
    return value_list

  return value


def get_env():
  """
  <Purpose>
    Parse environment for variables with prefix `ENV_PREFIX` and return
    a dict of key-value pairs.

    The prefix `ENV_PREFIX` is stripped from the keys in the returned dict.

    Values that contain colons (:) are split at the postion of the colons and
    converted into a list.


    Example:

    ```
    # Exporting variables in e.g. bash
    export IN_TOTO_ARTIFACT_BASE_PATH='/home/user/project'
    export IN_TOTO_ARTIFACT_EXCLUDE_PATTERNS='*.link:.gitignore'
    export IN_TOTO_LINK_CMD_EXEC_TIMEOUT='10'
    ```

    produces

    ```
    {
      "ARTIFACT_BASE_PATH": "/home/user/project"
      "ARTIFACT_EXCLUDE_PATTERNS": ["*.link", ".gitignore"]
      "LINK_CMD_EXEC_TIMEOUT": "10"
    }
    ```

  <Exceptions>
    None.

  <Side Effects>
    Calls function to read files from disk.

  <Returns>
    A dictionary containing the parsed key-value pairs.

  """
  env_dict = {}

  for name, value in os.environ.items():
    if (name.startswith(ENV_PREFIX) and
        len(name) > len(ENV_PREFIX)):
      stripped_name = name[len(ENV_PREFIX):]

      env_dict[stripped_name] = _colon_split(value)

  return env_dict


def get_rc():
  """
  <Purpose>
    Reads RCfiles from the paths defined in `RC_PATHS` and returns
    a dictionary with all parsed key-value pairs.

    The RCfile format is as expected by Python's builtin `ConfigParser` with
    the addition that values that contain colons (:) are split at the position
    of the colons and converted into a list.

    Section titles in RCfiles are ignored when parsing the key-value pairs.
    However, there has to be at least one section defined.

    The paths in `RC_PATHS` are ordered in reverse precedence, i.e. each file's
    settings override a previous file's settings, e.g. a setting defined
    in `.in_totorc` (in the current working dir) overrides the same
    setting defined in `~/.in_totorc` (in the user's home dir) and so on ...

    Example:

    ```
    # E.g. file `.in_totorc` in current working directory
    [in-toto setting]
    ARTIFACT_BASE_PATH = /home/user/project
    ARTIFACT_EXCLUDE_PATTERNS = *.link:.gitignore
    LINK_CMD_EXEC_TIMEOUT = 10
    ```

    produces

    ```
    {
      "ARTIFACT_BASE_PATH": "/home/user/project"
      "ARTIFACT_EXCLUDE_PATTERNS": ["*.link", ".gitignore"]
      "LINK_CMD_EXEC_TIMEOUT": "10"
    }
    ```

  <Exceptions>
    None.

  <Side Effects>
    Calls function to read files from disk.

  <Returns>
    A dictionary containing the parsed key-value pairs.

  """
  rc_dict = {}

  config = configparser.ConfigParser()
  # Reset `optionxform`'s default case conversion to enable case-sensitivity
  config.optionxform = str
  config.read(RC_PATHS)

  for section in config.sections():
    for name, value in config.items(section):
      rc_dict[name] = _colon_split(value)

  return rc_dict


def set_settings():
  """
  <Purpose>
    Calls functions that read in-toto related environment variables and RCfiles
    and overrides variables in `settings.py` with the retrieved values, if they
    are whitelisted in `IN_TOTO_SETTINGS`.

    Settings defined in RCfiles take precedence over settings defined in
    environment variables.

  <Exceptions>
    None.

  <Side Effects>
    Calls functions that read environment variables and files from disk.

  <Returns>
    None.

  """
  user_settings = get_env()
  user_settings.update(get_rc())

  # If the user has specified one of the settings whitelisted in
  # IN_TOTO_SETTINGS per envvar or rcfile, override the item in `settings.py`
  for setting in IN_TOTO_SETTINGS:
    user_setting = user_settings.get(setting)
    if user_setting:
      LOG.info("Setting (user): {0}={1}".format(
          setting, user_setting))
      setattr(in_toto.settings, setting, user_setting)

    else:
      default_setting = getattr(in_toto.settings, setting)
      LOG.info("Setting (default): {0}={1}".format(
          setting, default_setting))



#!/usr/bin/env python

# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Program Name>
  in_toto_keygen.py

<Author>
  Lukas Puehringer <lukas.puehringer@nyu.edu>
  Sachit Malik <i.sachitmalik@gmail.com>

<Started>
  June 28, 2017

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  A CLI tool for creating key files, and dumping them with <filename>,
  <filename>.pub for private and public keys respectively. It also takes an
  integer as an input, which specifies the length of the RSA key to be
  generated. By default it is set as 3072.

<Return Codes>
  2 if an exception occurred during argument parsing
  1 if an exception occurred
  0 if no exception occurred

"""
import sys
import argparse
import logging

from in_toto.common_args import title_case_action_groups
from in_toto import (
    __version__, SUPPORTED_KEY_TYPES, KEY_TYPE_RSA, KEY_TYPE_ED25519)
from securesystemslib import interface

# Command line interfaces should use in_toto base logger (c.f. in_toto.log)
LOG = logging.getLogger("in_toto")


def create_parser():
  """
  <Purpose>
    A function which parses the user supplied arguments.

  <Arguments>
    None

  <Exceptions>
    None

  <Returns>
    Parsed arguments (args object)
  """
  parser = argparse.ArgumentParser(
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description="in-toto-keygen is a tool to generate, optionally encrypt, and"
                " write cryptographic keys to disk. These keys may be used"
                " with other in-toto tooling to e.g. sign or verify link or"
                " layout metadata.")

  parser.epilog = """EXAMPLE USAGE

Generate RSA key pair of size 2048 bits, prompt for a password to encrypt
the private key, and write 'alice' (private encrypted) and 'alice.pub' (public)
as PEM-formatted key files to the current working directory.

  in-toto-keygen -p -t rsa -b 2048 alice


Generate unencrypted ed25519 key pair and write 'bob' (private) and 'bob.pub'
(public) as securesystemslib/json-formatted key files to the current working
directory.

  in-toto-keygen -t ed25519 bob


"""

  parser.add_argument("-p", "--prompt", action="store_true",
                            help="prompts for a password used to encrypt the"
                            " private key before storing it")

  parser.add_argument("-t", "--type", type=str,
                            choices=SUPPORTED_KEY_TYPES,
                            default=KEY_TYPE_RSA,
                            help="type of the key to be generated. '{rsa}'"
                            " keys are written in a 'PEM' format and"
                            " '{ed25519}' in a custom 'securesystemslib/json'"
                            " format. Default is '{rsa}'.".format(
                            rsa=KEY_TYPE_RSA, ed25519=KEY_TYPE_ED25519))


  parser.add_argument("name", type=str, metavar="<filename>",
                            help="filename for the resulting key files, which"
                            " are written to '<filename>' (private key) and"
                            " '<filename>.pub' (public key).")

  parser.add_argument("-b", "--bits", default=3072, type=int, metavar="<bits>",
                            help="key size, or key length, of the RSA key")

  parser.add_argument('--version', action='version',
                      version='{} {}'.format(parser.prog, __version__))

  title_case_action_groups(parser)

  return parser



def main():
  """
  First calls parse_args to parse the arguments, and then calls either
  _generate_and_write_rsa_keypair or _generate_and_write_ed25519_keypair
  depending upon the arguments. It then dumps the corresponding key files as:
  <filename> and <filename>.pub (Private key and Public key respectively)
  """
  parser = create_parser()
  args = parser.parse_args()

  try:
    if args.type == KEY_TYPE_RSA:
      interface._generate_and_write_rsa_keypair( # pylint: disable=protected-access
          filepath=args.name, bits=args.bits, prompt=args.prompt)
    elif args.type == KEY_TYPE_ED25519:
      interface._generate_and_write_ed25519_keypair( # pylint: disable=protected-access
          filepath=args.name, prompt=args.prompt)
    else:  # pragma: no cover
      LOG.error(
          "(in-toto-keygen) Unsupported keytype: {0}".format(str(args.type)))
      sys.exit(1)
    sys.exit(0)

  except Exception as e:
    LOG.error("(in-toto-keygen) {0}: {1}".format(type(e).__name__, e))
    sys.exit(1)

if __name__ == "__main__":
  main()



# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Program Name>
  log.py

<Author>
  Lukas Puehringer <lukas.puehringer@nyu.edu>

<Started>
  Oct 4, 2016

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Configures "in_toto" base logger, which can be used for debugging and user
  feedback in in-toto command line interfaces and libraries.

  Logging methods and levels are available through Python's logging module.

  If the log level is set to 'logging.DEBUG' log messages include
  additional information about the log statement. Moreover, calls to the
  `error` method will also output a stacktrace (if available).
  In all other log levels only the log message is shown without additional
  info.

  The default log level of the base logger is 'logging.WARNING', unless
  'in_toto.settings.DEBUG' is 'True', in that case the default log level is
  'logging.DEBUG'.

  The default handler of the base logger is a 'StreamHandler', which writes all
  log messages permitted by the used log level to 'sys.stderr'.


<Usage>
  This module should be imported in '__init__.py' to configure the base logger.
  Subsequently, command line interfaces should fetch the base logger by name
  and customize the log level according to any passed command line arguments.
  They can use the convenience method `setLevelVerboseOrQuiet` available on
  the custom base logger, e.g.:

  ```
  import logging
  LOG = logging.getLogger("in_toto")

  # parse args ...

  LOG.setLevelVerboseOrQuiet(args.verbose, args.quiet)

  # call library function ...
  ```

  Library modules can then create loggers, passing the module name, which will
  inherit the base logger's log level and format, e.g.:

  ```
  import logging
  LOG = logging.getLogger(__name__)

  LOG.warning("Shown per default.")

  LOG.info("Only shown if, log level was set to <= logging.INFO, e.g. in cli.")

  LOG.info("In debug mode it looks something like this)
  # in_toto.runlib:400:INFO:In debug mode it looks something like this

  ```

"""
import sys
import logging
import in_toto.settings

# Different log message formats for different log levels
FORMAT_MESSAGE = "%(message)s"
FORMAT_DEBUG = "%(name)s:%(lineno)d:%(levelname)s:%(message)s"

# Cache default logger class, should be logging.Logger if not changed elsewhere
_LOGGER_CLASS = logging.getLoggerClass()

# Create logger subclass
class InTotoLogger(_LOGGER_CLASS):
  """logger.Logging subclass, using providing custom error method and
  convenience method for log levels. """

  QUIET = logging.CRITICAL + 1

  def error(self, msg):
    """Show stacktrace depending on its availability and the logger's log
    level, i.e. only show stacktrace in DEBUG level. """
    show_stacktrace = (self.level == logging.DEBUG and
        sys.exc_info() != (None, None, None))
    return super(InTotoLogger, self).error(msg, exc_info=show_stacktrace)

  # Allow non snake_case function name for consistency with logging library
  def setLevelVerboseOrQuiet(self, verbose, quiet): # pylint: disable=invalid-name
    """Convenience method to set the logger's verbosity level based on the
    passed booleans verbose and quiet (useful for cli tools). """
    if verbose:
      self.setLevel(logging.INFO)

    elif quiet:
      # TODO: Is it enough to use logging.CRITICAL + 1 to suppress all output?
      # A saver way would be to use a NullHandler or Filters.
      self.setLevel(self.QUIET)


# Temporarily change logger default class to instantiate an in-toto base logger
logging.setLoggerClass(InTotoLogger)
LOGGER = logging.getLogger("in_toto")
logging.setLoggerClass(_LOGGER_CLASS)

# In DEBUG mode we log all log types and add additional information,
# otherwise we only log warning, error and critical and only the message.
if in_toto.settings.DEBUG: # pragma: no cover
  LEVEL = logging.DEBUG
  FORMAT_STRING = FORMAT_DEBUG

else:
  LEVEL = logging.WARNING
  FORMAT_STRING = FORMAT_MESSAGE

# Add a StreamHandler with the chosen format to in-toto's base logger,
# which will write log messages to `sys.stderr`.
FORMATTER = logging.Formatter(FORMAT_STRING)
HANDLER = logging.StreamHandler()
HANDLER.setFormatter(FORMATTER)
LOGGER.addHandler(HANDLER)
LOGGER.setLevel(LEVEL)



# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Program Name>
  runlib.py

<Author>
  Lukas Puehringer <lukas.puehringer@nyu.edu>

<Started>
  June 27, 2016

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides a wrapper for any command of the software supply chain.

  The wrapper performs the following tasks which are implemented in this
  library.

    - Record state of material (files the command is executed on)
    - Execute command
      - Capture stdout/stderr/return value of the executed command
    - Record state of product (files after the command was executed)
    - Return Metablock containing a Link object which can be can be signed
      and stored to disk
"""
import glob
import logging
import os
import itertools

from pathspec import PathSpec

import in_toto.settings
import in_toto.exceptions

from in_toto.models.link import (UNFINISHED_FILENAME_FORMAT, FILENAME_FORMAT,
    FILENAME_FORMAT_SHORT, UNFINISHED_FILENAME_FORMAT_GLOB)

import securesystemslib.formats
import securesystemslib.hash
import securesystemslib.exceptions
import securesystemslib.process
import securesystemslib.gpg

from in_toto.models.metadata import Metablock


# Inherits from in_toto base logger (c.f. in_toto.log)
LOG = logging.getLogger(__name__)





def _hash_artifact(filepath, hash_algorithms=None,
      normalize_line_endings=False):
  """Internal helper that takes a filename and hashes the respective file's
  contents using the passed hash_algorithms and returns a hashdict conformant
  with securesystemslib.formats.HASHDICT_SCHEMA. """
  if not hash_algorithms:
    hash_algorithms = ['sha256']

  securesystemslib.formats.HASHALGORITHMS_SCHEMA.check_match(hash_algorithms)
  hash_dict = {}

  for algorithm in hash_algorithms:
    digest_object = securesystemslib.hash.digest_filename(filepath, algorithm,
        normalize_line_endings=normalize_line_endings)
    hash_dict.update({algorithm: digest_object.hexdigest()})

  securesystemslib.formats.HASHDICT_SCHEMA.check_match(hash_dict)

  return hash_dict


def _apply_exclude_patterns(names, exclude_filter):
  """Exclude matched patterns from passed names."""
  included = set(names)

  # Assume old way for easier testing
  if hasattr(exclude_filter, '__iter__'):
    exclude_filter = PathSpec.from_lines('gitwildmatch', exclude_filter)

  for excluded in exclude_filter.match_files(names):
    included.discard(excluded)

  return sorted(included)


def _apply_left_strip(artifact_filepath, artifacts_dict, lstrip_paths=None):
  """ Internal helper function to left strip dictionary keys based on
  prefixes passed by the user. """
  if lstrip_paths:
    # If a prefix is passed using the argument --lstrip-paths,
    # that prefix is left stripped from the filepath passed.
    # Note: if the prefix doesn't include a trailing /, the dictionary key
    # may include an unexpected /.
    for prefix in lstrip_paths:
      if artifact_filepath.startswith(prefix):
        artifact_filepath = artifact_filepath[len(prefix):]
        break

    if artifact_filepath in artifacts_dict:
      raise in_toto.exceptions.PrefixError("Prefix selection has "
          "resulted in non unique dictionary key '{}'"
          .format(artifact_filepath))

  return artifact_filepath


def record_artifacts_as_dict(artifacts, exclude_patterns=None,
    base_path=None, follow_symlink_dirs=False, normalize_line_endings=False,
    lstrip_paths=None):
  """
  <Purpose>
    Hashes each file in the passed path list. If the path list contains
    paths to directories the directory tree(s) are traversed.

    The files a link command is executed on are called materials.
    The files that result form a link command execution are called
    products.

    Paths are normalized for matching and storing by left stripping "./"

    NOTE on exclude patterns:
      - Uses PathSpec to compile gitignore-style patterns, making use of the
        GitWildMatchPattern class (registered as 'gitwildmatch')

      - Patterns are checked for match against the full path relative to each
        path passed in the artifacts list

      - If a directory is excluded, all its files and subdirectories are also
        excluded

      - How it differs from .gitignore
            - No need to escape #
            - No ignoring of trailing spaces
            - No general negation with exclamation mark !
            - No special treatment of slash /
            - No special treatment of consecutive asterisks **

      - Exclude patterns are likely to become command line arguments or part of
        a config file.

  <Arguments>
    artifacts:
            A list of file or directory paths used as materials or products for
            the link command.

    exclude_patterns: (optional)
            Artifacts matched by the pattern are excluded from the result.
            Exclude patterns can be passed as argument or specified via
            ARTIFACT_EXCLUDE_PATTERNS setting (see `in_toto.settings`) or
            via envvars or rcfiles (see `in_toto.user_settings`).
            If passed, patterns specified via settings are overriden.

    base_path: (optional)
            Change to base_path and record artifacts relative from there.
            If not passed, current working directory is used as base_path.
            NOTE: The base_path part of the recorded artifact is not included
            in the returned paths.

    follow_symlink_dirs: (optional)
            Follow symlinked dirs if the linked dir exists (default is False).
            The recorded path contains the symlink name, not the resolved name.
            NOTE: This parameter toggles following linked directories only,
            linked files are always recorded, independently of this parameter.
            NOTE: Beware of infinite recursions that can occur if a symlink
            points to a parent directory or itself.

    normalize_line_endings: (optional)
            If True, replaces windows and mac line endings with unix line
            endings before hashing the content of the passed files, for
            cross-platform support.

    lstrip_paths: (optional)
            If a prefix path is passed, the prefix is left stripped from
            the path of every artifact that contains the prefix.

  <Exceptions>
    in_toto.exceptions.ValueError,
        if we cannot change to base path directory

    in_toto.exceptions.FormatError,
        if the list of exlcude patterns does not match format
        securesystemslib.formats.NAMES_SCHEMA

  <Side Effects>
    Calls functions to generate cryptographic hashes.

  <Returns>
    A dictionary with file paths as keys and the files' hashes as values.
  """

  artifacts_dict = {}

  if not artifacts:
    return artifacts_dict

  if base_path:
    LOG.info("Overriding setting ARTIFACT_BASE_PATH with passed"
        " base path.")
  else:
    base_path = in_toto.settings.ARTIFACT_BASE_PATH


  # Temporarily change into base path dir if set
  if base_path:
    original_cwd = os.getcwd()
    try:
      os.chdir(base_path)

    except Exception as e:
      raise ValueError("Could not use '{}' as base path: '{}'".format(
          base_path, e)) from  e

  # Normalize passed paths
  norm_artifacts = []
  for path in artifacts:
    norm_artifacts.append(os.path.normpath(path))

  # Passed exclude patterns take precedence over exclude pattern settings
  if exclude_patterns:
    LOG.info("Overriding setting ARTIFACT_EXCLUDE_PATTERNS with passed"
        " exclude patterns.")
  else:
    # TODO: Do we want to keep the exclude pattern setting?
    exclude_patterns = in_toto.settings.ARTIFACT_EXCLUDE_PATTERNS

  # Apply exclude patterns on the passed artifact paths if available
  if exclude_patterns:
    securesystemslib.formats.NAMES_SCHEMA.check_match(exclude_patterns)
    norm_artifacts = _apply_exclude_patterns(norm_artifacts, exclude_patterns)

  # Check if any of the prefixes passed for left stripping is a left substring
  # of another
  if lstrip_paths:
    for prefix_one, prefix_two in itertools.combinations(lstrip_paths, 2):
      if prefix_one.startswith(prefix_two) or \
          prefix_two.startswith(prefix_one):
        raise in_toto.exceptions.PrefixError("'{}' and '{}' "
            "triggered a left substring error".format(prefix_one, prefix_two))

  # Compile the gitignore-style patterns
  exclude_filter = PathSpec.from_lines('gitwildmatch', exclude_patterns or [])

  # Iterate over remaining normalized artifact paths
  for artifact in norm_artifacts:
    if os.path.isfile(artifact):
      # FIXME: this is necessary to provide consisency between windows
      # filepaths and *nix filepaths. A better solution may be in order
      # though...
      artifact = artifact.replace('\\', '/')
      key = _apply_left_strip(artifact, artifacts_dict, lstrip_paths)
      artifacts_dict[key] = _hash_artifact(artifact,
          normalize_line_endings=normalize_line_endings)

    elif os.path.isdir(artifact):
      for root, dirs, files in os.walk(artifact,
          followlinks=follow_symlink_dirs):
        # Create a list of normalized dirpaths
        dirpaths = []
        for dirname in dirs:
          norm_dirpath = os.path.normpath(os.path.join(root, dirname))
          dirpaths.append(norm_dirpath)

        # Applying exclude patterns on the directory paths returned by walk
        # allows to exclude a subdirectory 'sub' with a pattern 'sub'.
        # If we only applied the patterns below on the subdirectory's
        # containing file paths, we'd have to use a wildcard, e.g.: 'sub*'
        if exclude_patterns:
          dirpaths = _apply_exclude_patterns(dirpaths, exclude_filter)

        # Reset and refill dirs with remaining names after exclusion
        # Modify (not reassign) dirnames to only recurse into remaining dirs
        dirs[:] = []
        for dirpath in dirpaths:
          # Dirs only contain the basename and not the full path
          name = os.path.basename(dirpath)
          dirs.append(name)

        # Create a list of normalized filepaths
        filepaths = []
        for filename in files:
          norm_filepath = os.path.normpath(os.path.join(root, filename))

          # `os.walk` could also list dead symlinks, which would
          # result in an error later when trying to read the file
          if os.path.isfile(norm_filepath):
            filepaths.append(norm_filepath)

          else:
            LOG.info("File '{}' appears to be a broken symlink. Skipping..."
                .format(norm_filepath))

        # Apply exlcude patterns on the normalized file paths returned by walk
        if exclude_patterns:
          filepaths = _apply_exclude_patterns(filepaths, exclude_filter)

        for filepath in filepaths:
          # FIXME: this is necessary to provide consisency between windows
          # filepaths and *nix filepaths. A better solution may be in order
          # though...
          normalized_filepath = filepath.replace("\\", "/")
          key = _apply_left_strip(
              normalized_filepath, artifacts_dict, lstrip_paths)
          artifacts_dict[key] = _hash_artifact(filepath,
              normalize_line_endings=normalize_line_endings)

    # Path is no file and no directory
    else:
      LOG.info("path: {} does not exist, skipping..".format(artifact))


  # Change back to where original current working dir
  if base_path:
    os.chdir(original_cwd)

  return artifacts_dict

def execute_link(link_cmd_args, record_streams):
  """
  <Purpose>
    Executes the passed command plus arguments in a subprocess and returns
    the return value of the executed command. If the specified standard output
    and standard error of the command are recorded and also returned to the
    caller.

  <Arguments>
    link_cmd_args:
            A list where the first element is a command and the remaining
            elements are arguments passed to that command.
    record_streams:
            A bool that specifies whether to redirect standard output and
            and standard error to a temporary file which is returned to the
            caller (True) or not (False).

  <Exceptions>
    OSError:
            The given command is not present or non-executable

    securesystemslib.process.subprocess.TimeoutExpired:
            The execution of the given command times out. The default timeout
            is securesystemslib.settings.SUBPROCESS_TIMEOUT.


  <Side Effects>
    Executes passed command in a subprocess and redirects stdout and stderr
    if specified.

  <Returns>
    - A dictionary containing standard output and standard error of the
      executed command, called by-products.
      Note: If record_streams is False, the dict values are empty strings.
    - The return value of the executed command.
  """
  if record_streams:
    return_code, stdout_str, stderr_str = \
        securesystemslib.process.run_duplicate_streams(link_cmd_args,
            timeout=float(in_toto.settings.LINK_CMD_EXEC_TIMEOUT))

  else:
    process = securesystemslib.process.run(link_cmd_args, check=False,
      timeout=float(in_toto.settings.LINK_CMD_EXEC_TIMEOUT),
      stdout=securesystemslib.process.DEVNULL,
      stderr=securesystemslib.process.DEVNULL)
    stdout_str = stderr_str = ""
    return_code = process.returncode

  return {
      "stdout": stdout_str,
      "stderr": stderr_str,
      "return-value": return_code
    }


def in_toto_mock(name, link_cmd_args):
  """
  <Purpose>
    in_toto_run with defaults
     - Records materials and products in current directory
     - Does not sign resulting link file
     - Stores resulting link file under "<name>.link"

  <Arguments>
    name:
            A unique name to relate mock link metadata with a step or
            inspection defined in the layout.
    link_cmd_args:
            A list where the first element is a command and the remaining
            elements are arguments passed to that command.

  <Exceptions>
    None.

  <Side Effects>
    Writes newly created link metadata file to disk using the filename scheme
    from link.FILENAME_FORMAT_SHORT

  <Returns>
    Newly created Metablock object containing a Link object

  """
  link_metadata = in_toto_run(name, ["."], ["."], link_cmd_args,
      record_streams=True)

  filename = FILENAME_FORMAT_SHORT.format(step_name=name)
  LOG.info("Storing unsigned link metadata to '{}'...".format(filename))
  link_metadata.dump(filename)
  return link_metadata


def _check_match_signing_key(signing_key):
  """ Helper method to check if the signing_key has securesystemslib's
  KEY_SCHEMA and the private part is not empty.
  # FIXME: Add private key format check to formats
  """
  securesystemslib.formats.KEY_SCHEMA.check_match(signing_key)
  if not signing_key["keyval"].get("private"):
    raise securesystemslib.exceptions.FormatError(
        "Signing key needs to be a private key.")


def in_toto_run(name, material_list, product_list, link_cmd_args,
    record_streams=False, signing_key=None, gpg_keyid=None,
    gpg_use_default=False, gpg_home=None, exclude_patterns=None,
    base_path=None, compact_json=False, record_environment=False,
    normalize_line_endings=False, lstrip_paths=None, metadata_directory=None):
  """Performs a supply chain step or inspection generating link metadata.

  Executes link_cmd_args, recording paths and hashes of files before and after
  command execution (aka. artifacts) in a link metadata file. The metadata is
  signed with the passed signing_key, a gpg key identified by its ID, or the
  default gpg key. If multiple key arguments are passed, only one key is used
  in above order of precedence. The resulting link file is written to
  ``STEP-NAME.KEYID-PREFIX.link``. If no key argument is passed the link
  metadata is neither signed nor written to disk.

  Arguments:
    name: A unique name to associate link metadata with a step or inspection.

    material_list: A list of artifact paths to be recorded before command
        execution. Directories are traversed recursively.

    product_list: A list of artifact paths to be recorded after command
        execution. Directories are traversed recursively.

    link_cmd_args: A list where the first element is a command and the
        remaining elements are arguments passed to that command.

    record_streams (optional): A boolean indicating if standard output and
        standard error of the link command should be recorded in the link
        metadata in addition to being displayed while the command is executed.

    signing_key (optional): A key used to sign the resulting link metadata. The
        format is securesystemslib.formats.KEY_SCHEMA.

    gpg_keyid (optional): A keyid used to identify a local gpg key used to sign
        the resulting link metadata.

    gpg_use_default (optional): A boolean indicating if the default gpg key
        should be used to sign the resulting link metadata.

    gpg_home (optional): A path to the gpg home directory. If not set the
        default gpg home directory is used.

    exclude_patterns (optional): A list of filename patterns to exclude certain
        files from being recorded as artifacts. See Config docs for details.

    base_path (optional): A path relative to which artifacts are recorded.
        Default is the current working directory.

    compact_json (optional): A boolean indicating if the resulting link
        metadata should be written in the most compact JSON representation.

    record_environment (optional): A boolean indicating if information about
        the environment should be added in the resulting link metadata.

    normalize_line_endings (optional): A boolean indicating if line endings of
        artifacts should be normalized before hashing for cross-platform
        support.

    lstrip_paths (optional): A list of path prefixes used to left-strip
        artifact paths before storing them in the resulting link metadata.

    metadata_directory (optional): A directory path to write the resulting link
        metadata file to. Default destination is the current working directory.

  Raises:
    securesystemslib.exceptions.FormatError: Passed arguments are malformed.

    ValueError: Cannot change to base path directory.

    securesystemslib.exceptions.StorageError: Cannot hash artifacts.

    PrefixError: Left-stripping artifact paths results in non-unique dict keys.

    securesystemslib.process.subprocess.TimeoutExpired: Link command times out.

    IOError, FileNotFoundError, NotADirectoryError, PermissionError:
        Cannot write link metadata.

    securesystemslib.exceptions.CryptoError, \
            securesystemslib.exceptions.UnsupportedAlgorithmError:
        Signing errors.

    ValueError, OSError, securesystemslib.gpg.exceptions.CommandError, \
            securesystemslib.gpg.exceptions.KeyNotFoundError:
        gpg signing errors.

  Side Effects:
    Reads artifact files from disk.
    Runs link command in subprocess.
    Calls system gpg in a subprocess, if a gpg key argument is passed.
    Writes link metadata file to disk, if any key argument is passed.

  Returns:
    A Metablock object that contains the resulting link object.

  """
  LOG.info("Running '{}'...".format(name))

  # Check key formats to fail early
  if signing_key:
    _check_match_signing_key(signing_key)
  if gpg_keyid:
    securesystemslib.formats.KEYID_SCHEMA.check_match(gpg_keyid)

  if exclude_patterns:
    securesystemslib.formats.NAMES_SCHEMA.check_match(exclude_patterns)

  if base_path:
    securesystemslib.formats.PATH_SCHEMA.check_match(base_path)

  if metadata_directory:
    securesystemslib.formats.PATH_SCHEMA.check_match(metadata_directory)

  if material_list:
    LOG.info("Recording materials '{}'...".format(", ".join(material_list)))

  materials_dict = record_artifacts_as_dict(material_list,
      exclude_patterns=exclude_patterns, base_path=base_path,
      follow_symlink_dirs=True, normalize_line_endings=normalize_line_endings,
      lstrip_paths=lstrip_paths)

  if link_cmd_args:
    LOG.info("Running command '{}'...".format(" ".join(link_cmd_args)))
    byproducts = execute_link(link_cmd_args, record_streams)
  else:
    byproducts = {}

  if product_list:
    securesystemslib.formats.PATHS_SCHEMA.check_match(product_list)
    LOG.info("Recording products '{}'...".format(", ".join(product_list)))

  products_dict = record_artifacts_as_dict(product_list,
      exclude_patterns=exclude_patterns, base_path=base_path,
      follow_symlink_dirs=True, normalize_line_endings=normalize_line_endings,
      lstrip_paths=lstrip_paths)

  LOG.info("Creating link metadata...")
  environment = {}
  if record_environment:
    environment['workdir'] = os.getcwd().replace('\\', '/')

  link = in_toto.models.link.Link(name=name,
      materials=materials_dict, products=products_dict, command=link_cmd_args,
      byproducts=byproducts, environment=environment)

  link_metadata = Metablock(signed=link, compact_json=compact_json)

  signature = None
  if signing_key:
    LOG.info("Signing link metadata using passed key...")
    signature = link_metadata.sign(signing_key)

  elif gpg_keyid:
    LOG.info("Signing link metadata using passed GPG keyid...")
    signature = link_metadata.sign_gpg(gpg_keyid, gpg_home=gpg_home)

  elif gpg_use_default:
    LOG.info("Signing link metadata using default GPG key ...")
    signature = link_metadata.sign_gpg(gpg_keyid=None, gpg_home=gpg_home)

  # We need the signature's keyid to write the link to keyid infix'ed filename
  if signature:
    signing_keyid = signature["keyid"]
    filename = FILENAME_FORMAT.format(step_name=name, keyid=signing_keyid)

    if metadata_directory is not None:
      filename = os.path.join(metadata_directory, filename)

    LOG.info("Storing link metadata to '{}'...".format(filename))
    link_metadata.dump(filename)

  return link_metadata


def in_toto_record_start(step_name, material_list, signing_key=None,
    gpg_keyid=None, gpg_use_default=False, gpg_home=None,
    exclude_patterns=None, base_path=None, record_environment=False,
    normalize_line_endings=False, lstrip_paths=None):
  """Generates preliminary link metadata.

  Records paths and hashes of materials in a preliminary link metadata file.
  The metadata is signed with the passed signing_key, a gpg key identified by
  its ID, or the default gpg key. If multiple key arguments are passed, only
  one key is used in above order of precedence. At least one key argument must
  be passed. The resulting link file is written to
  ``.STEP-NAME.KEYID-PREFIX.link-unfinished``.

  Use this function together with in_toto_record_stop as an alternative to
  in_toto_run, in order to provide evidence for supply chain steps that cannot
  be carried out by a single command.

  Arguments:
    step_name: A unique name to associate link metadata with a step.

    material_list: A list of artifact paths to be recorded as materials.
        Directories are traversed recursively.

    signing_key (optional): A key used to sign the resulting link metadata. The
        format is securesystemslib.formats.KEY_SCHEMA.

    gpg_keyid (optional): A keyid used to identify a local gpg key used to sign
        the resulting link metadata.

    gpg_use_default (optional): A boolean indicating if the default gpg key
        should be used to sign the resulting link metadata.

    gpg_home (optional): A path to the gpg home directory. If not set the
        default gpg home directory is used.

    exclude_patterns (optional): A list of filename patterns to exclude certain
        files from being recorded as artifacts. See Config docs for details.

    base_path (optional): A path relative to which artifacts are recorded.
        Default is the current working directory.

    record_environment (optional): A boolean indicating if information about
        the environment should be added in the resulting link metadata.

    normalize_line_endings (optional): A boolean indicating if line endings of
        artifacts should be normalized before hashing for cross-platform
        support.

    lstrip_paths (optional): A list of path prefixes used to left-strip
        artifact paths before storing them in the resulting link metadata.

  Raises:
    securesystemslib.exceptions.FormatError: Passed arguments are malformed.

    ValueError: None of signing_key, gpg_keyid or gpg_use_default=True is
        passed.

    securesystemslib.exceptions.StorageError: Cannot hash artifacts.

    PrefixError: Left-stripping artifact paths results in non-unique dict keys.

    securesystemslib.process.subprocess.TimeoutExpired: Link command times out.

    IOError, PermissionError:
        Cannot write link metadata.

    securesystemslib.exceptions.CryptoError, \
            securesystemslib.exceptions.UnsupportedAlgorithmError:
        Signing errors.

    ValueError, OSError, securesystemslib.gpg.exceptions.CommandError, \
            securesystemslib.gpg.exceptions.KeyNotFoundError:
        gpg signing errors.

  Side Effects:
    Reads artifact files from disk.
    Calls system gpg in a subprocess, if a gpg key argument is passed.
    Writes preliminary link metadata file to disk.

  """
  LOG.info("Start recording '{}'...".format(step_name))

  # Fail if there is no signing key arg at all
  if not signing_key and not gpg_keyid and not gpg_use_default:
    raise ValueError("Pass either a signing key, a gpg keyid or set"
        " gpg_use_default to True!")

  # Check key formats to fail early
  if signing_key:
    _check_match_signing_key(signing_key)
  if gpg_keyid:
    securesystemslib.formats.KEYID_SCHEMA.check_match(gpg_keyid)

  if exclude_patterns:
    securesystemslib.formats.NAMES_SCHEMA.check_match(exclude_patterns)

  if base_path:
    securesystemslib.formats.PATH_SCHEMA.check_match(base_path)

  if material_list:
    LOG.info("Recording materials '{}'...".format(", ".join(material_list)))

  materials_dict = record_artifacts_as_dict(material_list,
      exclude_patterns=exclude_patterns, base_path=base_path,
      follow_symlink_dirs=True, normalize_line_endings=normalize_line_endings,
      lstrip_paths=lstrip_paths)

  LOG.info("Creating preliminary link metadata...")
  environment = {}
  if record_environment:
    environment['workdir'] = os.getcwd().replace('\\', '/')

  link = in_toto.models.link.Link(name=step_name,
          materials=materials_dict, products={}, command=[], byproducts={},
          environment=environment)

  link_metadata = Metablock(signed=link)

  if signing_key:
    LOG.info("Signing link metadata using passed key...")
    signature = link_metadata.sign(signing_key)

  elif gpg_keyid:
    LOG.info("Signing link metadata using passed GPG keyid...")
    signature = link_metadata.sign_gpg(gpg_keyid, gpg_home=gpg_home)

  else:  # (gpg_use_default)
    LOG.info("Signing link metadata using default GPG key ...")
    signature = link_metadata.sign_gpg(gpg_keyid=None, gpg_home=gpg_home)

  # We need the signature's keyid to write the link to keyid infix'ed filename
  signing_keyid = signature["keyid"]

  unfinished_fn = UNFINISHED_FILENAME_FORMAT.format(step_name=step_name,
    keyid=signing_keyid)

  LOG.info(
      "Storing preliminary link metadata to '{}'...".format(unfinished_fn))
  link_metadata.dump(unfinished_fn)



def in_toto_record_stop(step_name, product_list, signing_key=None,
    gpg_keyid=None, gpg_use_default=False, gpg_home=None,
    exclude_patterns=None, base_path=None, normalize_line_endings=False,
    lstrip_paths=None, metadata_directory=None):
  """Finalizes preliminary link metadata generated with in_toto_record_start.

  Loads preliminary link metadata file, verifies its signature, and records
  paths and hashes as products, thus finalizing the link metadata. The metadata
  is signed with the passed signing_key, a gpg key identified by its ID, or the
  default gpg key. If multiple key arguments are passed, only one key is used
  in above order of precedence. At least one key argument must be passed and it
  must be the same as the one used to sign the preliminary link metadata file.
  The resulting link file is written to ``STEP-NAME.KEYID-PREFIX.link``.

  Use this function together with in_toto_record_start as an alternative to
  in_toto_run, in order to provide evidence for supply chain steps that cannot
  be carried out by a single command.

  Arguments:
    step_name: A unique name to associate link metadata with a step.

    product_list: A list of artifact paths to be recorded as products.
        Directories are traversed recursively.

    signing_key (optional): A key used to sign the resulting link metadata. The
        format is securesystemslib.formats.KEY_SCHEMA.

    gpg_keyid (optional): A keyid used to identify a local gpg key used to sign
        the resulting link metadata.

    gpg_use_default (optional): A boolean indicating if the default gpg key
        should be used to sign the resulting link metadata.

    gpg_home (optional): A path to the gpg home directory. If not set the
        default gpg home directory is used.

    exclude_patterns (optional): A list of filename patterns to exclude certain
        files from being recorded as artifacts.

    base_path (optional): A path relative to which artifacts are recorded.
        Default is the current working directory.

    normalize_line_endings (optional): A boolean indicating if line endings of
        artifacts should be normalized before hashing for cross-platform
        support.

    lstrip_paths (optional): A list of path prefixes used to left-strip
        artifact paths before storing them in the resulting link metadata.

    metadata_directory (optional): A directory path to write the resulting link
        metadata file to. Default destination is the current working directory.

  Raises:
    securesystemslib.exceptions.FormatError: Passed arguments are malformed.

    ValueError: None of signing_key, gpg_keyid or gpg_use_default=True is
        passed.

    LinkNotFoundError: No preliminary link metadata file found.

    securesystemslib.exceptions.StorageError: Cannot hash artifacts.

    PrefixError: Left-stripping artifact paths results in non-unique dict keys.

    securesystemslib.process.subprocess.TimeoutExpired: Link command times out.

    IOError, FileNotFoundError, NotADirectoryError, PermissionError:
        Cannot write link metadata.

    securesystemslib.exceptions.CryptoError, \
            securesystemslib.exceptions.UnsupportedAlgorithmError:
        Signing errors.

    ValueError, OSError, securesystemslib.gpg.exceptions.CommandError, \
            securesystemslib.gpg.exceptions.KeyNotFoundError:
        gpg signing errors.

  Side Effects:
    Reads preliminary link metadata file from disk.
    Reads artifact files from disk.
    Calls system gpg in a subprocess, if a gpg key argument is passed.
    Writes resulting link metadata file to disk.
    Removes preliminary link metadata file from disk.

  """
  LOG.info("Stop recording '{}'...".format(step_name))

  # Check that we have something to sign and if the formats are right
  if not signing_key and not gpg_keyid and not gpg_use_default:
    raise ValueError("Pass either a signing key, a gpg keyid or set"
        " gpg_use_default to True")

  if signing_key:
    _check_match_signing_key(signing_key)
  if gpg_keyid:
    securesystemslib.formats.KEYID_SCHEMA.check_match(gpg_keyid)

  if exclude_patterns:
    securesystemslib.formats.NAMES_SCHEMA.check_match(exclude_patterns)

  if base_path:
    securesystemslib.formats.PATH_SCHEMA.check_match(base_path)

  if metadata_directory:
    securesystemslib.formats.PATH_SCHEMA.check_match(metadata_directory)

  # Load preliminary link file
  # If we have a signing key we can use the keyid to construct the name
  if signing_key:
    unfinished_fn = UNFINISHED_FILENAME_FORMAT.format(step_name=step_name,
        keyid=signing_key["keyid"])

  # FIXME: Currently there is no way to know the default GPG key's keyid and
  # so we glob for preliminary link files
  else:
    unfinished_fn_glob = UNFINISHED_FILENAME_FORMAT_GLOB.format(
        step_name=step_name, pattern="*")
    unfinished_fn_list = glob.glob(unfinished_fn_glob)

    if not len(unfinished_fn_list):
      raise in_toto.exceptions.LinkNotFoundError("Could not find a preliminary"
          " link for step '{}' in the current working directory.".format(
          step_name))

    if len(unfinished_fn_list) > 1:
      raise in_toto.exceptions.LinkNotFoundError("Found more than one"
          " preliminary links for step '{}' in the current working directory:"
          " {}. We need exactly one to stop recording.".format(
          step_name, ", ".join(unfinished_fn_list)))

    unfinished_fn = unfinished_fn_list[0]

  LOG.info("Loading preliminary link metadata '{}'...".format(unfinished_fn))
  link_metadata = Metablock.load(unfinished_fn)

  # The file must have been signed by the same key
  # If we have a signing_key we use it for verification as well
  if signing_key:
    LOG.info(
        "Verifying preliminary link signature using passed signing key...")
    keyid = signing_key["keyid"]
    verification_key = signing_key

  elif gpg_keyid:
    LOG.info("Verifying preliminary link signature using passed gpg key...")
    gpg_pubkey = securesystemslib.gpg.functions.export_pubkey(
        gpg_keyid, gpg_home)
    keyid = gpg_pubkey["keyid"]
    verification_key = gpg_pubkey

  else: # must be gpg_use_default
    # FIXME: Currently there is no way to know the default GPG key's keyid
    # before signing. As a workaround we extract the keyid of the preliminary
    # Link file's signature and try to export a pubkey from the gpg
    # home directory. We do this even if a gpg_keyid was specified, because gpg
    # accepts many different ids (mail, name, parts of an id, ...) but we
    # need a specific format.
    LOG.info("Verifying preliminary link signature using default gpg key...")
    keyid = link_metadata.signatures[0]["keyid"]
    gpg_pubkey = securesystemslib.gpg.functions.export_pubkey(
        keyid, gpg_home)
    verification_key = gpg_pubkey

  link_metadata.verify_signature(verification_key)

  # Record products if a product path list was passed
  if product_list:
    LOG.info("Recording products '{}'...".format(", ".join(product_list)))

  link_metadata.signed.products = record_artifacts_as_dict(
      product_list, exclude_patterns=exclude_patterns, base_path=base_path,
      follow_symlink_dirs=True, normalize_line_endings=normalize_line_endings,
      lstrip_paths=lstrip_paths)

  link_metadata.signatures = []
  if signing_key:
    LOG.info("Updating signature with key '{:.8}...'...".format(keyid))
    link_metadata.sign(signing_key)

  else: # gpg_keyid or gpg_use_default
    # In both cases we use the keyid we got from verifying the preliminary
    # link signature above.
    LOG.info("Updating signature with gpg key '{:.8}...'...".format(keyid))
    link_metadata.sign_gpg(keyid, gpg_home)

  fn = FILENAME_FORMAT.format(step_name=step_name, keyid=keyid)

  if metadata_directory is not None:
    fn = os.path.join(metadata_directory, fn)

  LOG.info("Storing link metadata to '{}'...".format(fn))
  link_metadata.dump(fn)

  LOG.info("Removing unfinished link metadata '{}'...".format(unfinished_fn))
  os.remove(unfinished_fn)



# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Program Name>
  common.py

<Author>
  Lukas Puehringer <lukas.puehringer@nyu.edu>
  Santiago Torres <santiago@nyu.edu>

<Started>
  Sep 23, 2016

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides base classes for various classes in the model.

"""

import json
import attr
import inspect
import securesystemslib.formats



class ValidationMixin(object):
  """ The validation mixin provides a self-inspecting method, validate, to
  allow in-toto's objects to check that they are proper. """

  def validate(self):
    """Validates attributes of the instance.

    Raises:
      securesystemslib.formats.FormatError: An attribute value is invalid.

    """
    for method in inspect.getmembers(self, predicate=inspect.ismethod):
      if method[0].startswith("_validate_"):
        method[1]()



@attr.s(repr=False, init=False)
class Signable(ValidationMixin):
  """Objects with base class Signable are to be included in a Metablock class
  to be signed (hence the name). They provide a `signable_bytes` property
  used to create deterministic signatures. """

  def __repr__(self):
    """Returns an indented JSON string of the metadata object. """
    return json.dumps(attr.asdict(self),
        indent=1, separators=(",", ": "), sort_keys=True)

  @property
  def signable_bytes(self):
    """The UTF-8 encoded canonical JSON byte representation of the dictionary
    representation of the instance. """
    return securesystemslib.formats.encode_canonical(
        attr.asdict(self)).encode("UTF-8")



# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Program Name>
  verifylib.py

<Author>
  Lukas Puehringer <lukas.puehringer@nyu.edu>
  Santiago Torres-Arias <santiago@nyu.edu>

<Started>
  June 28, 2016

<Copyright>
  See LICENSE for licensing information.

<Purpose>

  Provides a library to verify an in-toto final product containing
  a software supply chain layout.

  Take a look at `in_toto_verify`'s docstring for more details about the
  entire verification workflow.

"""

import os
import datetime
import iso8601
import fnmatch
import logging
from dateutil import tz

import securesystemslib.exceptions

import in_toto.settings
import in_toto.runlib
import in_toto.models.layout
import in_toto.models.link
import in_toto.formats
from in_toto.models.metadata import Metablock
from in_toto.exceptions import (RuleVerificationError, LayoutExpiredError,
    ThresholdVerificationError, BadReturnValueError,
    SignatureVerificationError)
from securesystemslib.gpg.exceptions import KeyExpirationError
import in_toto.rulelib

# Inherits from in_toto base logger (c.f. in_toto.log)
LOG = logging.getLogger(__name__)

RULE_TRACE = {}


def _raise_on_bad_retval(return_value, command=None):
  """
  <Purpose>
    Internal function that checks return values of shell commands, e.g. from
    inspections. Raises exception if the passed value is non-int and non-zero.

  <Arguments>
    return_value:
            The return value to be verified
    command: (optional)
            The command whose execution returned the value, used for exception
            message.

  <Exceptions>
    BadReturnValueError if the return_value is non-int and non-zero

  <Side Effects>
    None.

  <Returns>
    None.
  """

  msg = "Got non-{what} " + "return value '{}'".format(return_value)
  if command:
    msg = "{0} from command '{1}'.".format(msg, command)
  else:
    msg = "{0}.".format(msg)

  if not isinstance(return_value, int):
    raise BadReturnValueError(msg.format(what="int"))

  # TODO: in-toto specification suggests special behavior on
  # return_value == 127, but does not fully define that behavior yet

  if return_value != 0:
    raise BadReturnValueError(msg.format(what="zero"))


def load_links_for_layout(layout, link_dir_path):
  """
  <Purpose>
    Try to load all existing metadata files for each Step of the Layout
    from the current directory.

    For each step the metadata might consist of multiple (thresholds) Link
    or Layout (sub-layouts) files.

  <Arguments>
    layout:
          Layout object

    link_dir_path:
          A path to directory where links are loaded from


  <Side Effects>
    Calls function to read files from disk

  <Exceptions>
    in_toto.exceptions.LinkNotFoundError,
            if fewer than `threshold` link files can be found for any given
            step of the supply chain (preliminary threshold check)

  <Returns>
    A dictionary carrying all the found metadata corresponding to the
    passed layout, e.g.:

    {
      <step name> : {
        <functionary key id> : <Metablock containing a Link or Layout object>,
        ...
      }, ...
    }


  """
  steps_metadata = {}

  # Iterate over all the steps in the layout
  for step in layout.steps:
    links_per_step = {}

    # We try to load a link for every authorized functionary, but don't fail
    # if the file does not exist (authorized != required)
    # FIXME: Should we really pass on IOError, or just skip inexistent links?
    for authorized_keyid in step.pubkeys:
      # Iterate over the authorized key and if present over subkeys
      for keyid in [authorized_keyid] + list(layout.keys.get(authorized_keyid,
          {}).get("subkeys", {}).keys()):

        filename = in_toto.models.link.FILENAME_FORMAT.format(
            step_name=step.name, keyid=keyid)
        filepath = os.path.join(link_dir_path, filename)

        try:
          metadata = Metablock.load(filepath)
          links_per_step[keyid] = metadata

        except IOError:
          pass

    # This is only a preliminary threshold check, based on (authorized)
    # filenames, to fail early. A more thorough signature-based threshold
    # check is indispensable.
    if len(links_per_step) < step.threshold:
      raise in_toto.exceptions.LinkNotFoundError("Step '{0}' requires '{1}'"
          " link metadata file(s), found '{2}'."
          .format(step.name, step.threshold, len(links_per_step)))

    steps_metadata[step.name] = links_per_step

  return steps_metadata


def run_all_inspections(layout, persist_inspection_links):
  """
  <Purpose>
    Extracts all inspections from a passed Layout's inspect field and
    iteratively runs each command defined in the Inspection's `run` field using
    `runlib.in_toto_run`, which returns a Metablock object containing a Link
    object.

    If a link command returns non-zero the verification is aborted.

  <Arguments>
    layout:
            A Layout object which is used to extract the Inspections.

    persist_inspection_links:
            A boolean indicating whether link metadata files for inspection
            are written to cwd.

  <Exceptions>
    Calls function that raises BadReturnValueError if an inspection returned
    non-int or non-zero.

  <Returns>
    A dictionary of metadata about the executed inspections, e.g.:

    {
      <inspection name> : {
        <Metablock containing a Link object>,
        ...
      }, ...
    }

  """
  inspection_links_dict = {}
  for inspection in layout.inspect:
    LOG.info("Executing command for inspection '{}'...".format(
        inspection.name))

    # FIXME: We don't want to use the base path for runlib so we patch this
    # for now. This will not stay!
    base_path_backup = in_toto.settings.ARTIFACT_BASE_PATH
    in_toto.settings.ARTIFACT_BASE_PATH = None

    # FIXME: What should we record as material/product?
    # Is the current directory a sensible default? In general?
    # If so, we should probably make it a default in run_link
    # We could use artifact rule paths.
    material_list = product_list = ["."]
    link = in_toto.runlib.in_toto_run(inspection.name, material_list,
        product_list, inspection.run)

    _raise_on_bad_retval(
        link.signed.byproducts.get("return-value"),
        inspection.run)

    inspection_links_dict[inspection.name] = link

    # If client requests persistent inspection links,
    # Dump the inspection link file for auditing
    # Keep in mind that this pollutes the verifier's (client's) filesystem.
    if persist_inspection_links:
      filename = in_toto.models.link.FILENAME_FORMAT_SHORT.format(
          step_name=inspection.name)
      link.dump(filename)

    in_toto.settings.ARTIFACT_BASE_PATH = base_path_backup

  return inspection_links_dict


def verify_layout_expiration(layout):
  """
  <Purpose>
    Raises an exception if the passed layout has expired, i.e. if its
    `expires` property is lesser "now".
    Time zone aware datetime objects in UTC+00:00 (Zulu Time) are used.

  <Arguments>
    layout:
            The Layout object to be verified.

  <Exceptions>
    LayoutExpiredError
    TBA (see https://github.com/in-toto/in-toto/issues/6)

  <Side Effects>
    None.

  """
  expire_datetime = iso8601.parse_date(layout.expires)
  if expire_datetime < datetime.datetime.now(tz.tzutc()):
    raise LayoutExpiredError("Layout expired")


def substitute_parameters(layout, parameter_dictionary):
  """
  <Purpose>
    This function is a transitionary measure for parameter substitution (or
    any other solution defined by the in-toto team). As of now, it acts as
    a very simple replacement layer for python-like parameters

  <Arguments>
    layout:
            The Layout object to process.

      parameter_dictionary:
            A dictionary containing key-value pairs for substitution.

  <Exceptions>
    securesystemslib.exceptions.FormatError:
      if the parameter dictionary is malformed.

    KeyError:
      if one of the keys in the parameter dictionary are not present for
      substitution

  <Side Effects>
    The layout object will have any tags replaced with the corresponding
    values defined in the parameter dictionary.
  """
  in_toto.formats.PARAMETER_DICTIONARY_SCHEMA.check_match(parameter_dictionary)

  for step in layout.steps:

    new_material_rules = []
    for rule in step.expected_materials:
      new_rule = []
      for stanza in rule:
        new_rule.append(stanza.format(**parameter_dictionary))
      new_material_rules.append(new_rule)

    new_product_rules = []
    for rule in step.expected_products:
      new_rule = []
      for stanza in rule:
        new_rule.append(stanza.format(**parameter_dictionary))
      new_product_rules.append(new_rule)

    new_expected_command = []
    for argv in step.expected_command:
      new_expected_command.append(argv.format(**parameter_dictionary))

    step.expected_command = new_expected_command
    step.expected_materials = new_material_rules
    step.expected_products = new_product_rules

  for inspection in layout.inspect:
    new_material_rules = []
    for rule in inspection.expected_materials:
      new_rule = []
      for stanza in rule:
        new_rule.append(stanza.format(**parameter_dictionary))
      new_material_rules.append(new_rule)

    new_product_rules = []
    for rule in inspection.expected_products:
      new_rule = []
      for stanza in rule:
        new_rule.append(stanza.format(**parameter_dictionary))
      new_product_rules.append(new_rule)

    new_run = []
    for argv in inspection.run:
      new_run.append(argv.format(**parameter_dictionary))

    inspection.run = new_run
    inspection.expected_materials = new_material_rules
    inspection.expected_products = new_product_rules



def verify_layout_signatures(layout_metablock, keys_dict):
  """
  <Purpose>
    Iteratively verifies the signatures of a Metablock object containing
    a Layout object for every verification key in the passed keys dictionary.

    Requires at least one key to be passed and requires every passed key to
    find a valid signature.

  <Arguments>
    layout_metablock:
            A Metablock object containing a Layout whose signatures are
            verified.

    keys_dict:
            A dictionary of keys to verify the signatures conformant with
            securesystemslib.formats.VERIFICATION_KEY_DICT_SCHEMA.

  <Exceptions>
    securesystemslib.exceptions.FormatError
      if the passed key dict does not match VERIFICATION_KEY_DICT_SCHEMA.

    SignatureVerificationError
      if an empty verification key dictionary was passed, or
      if any of the passed verification keys fails to verify a signature.

    securesystemslib.gpg.exceptions.KeyExpirationError:
      if any of the passed verification keys is an expired gpg key

  """
  securesystemslib.formats.VERIFICATION_KEY_DICT_SCHEMA.check_match(keys_dict)

  # Fail if an empty verification key dictionary was passed
  if len(keys_dict) < 1:
    raise SignatureVerificationError("Layout signature verification"
        " requires at least one key.")

  # Fail if any of the passed keys can't verify a signature on the Layout
  for junk, verify_key in keys_dict.items():
    layout_metablock.verify_signature(verify_key)


def verify_link_signature_thresholds(layout, chain_link_dict):
  """
  <Purpose>
    Verify that for each step of the layout there are at least `threshold`
    links, signed by different authorized functionaries and return the chain
    link dictionary containing only authorized links whose signatures
    were successfully verified.

    NOTE: If the layout's key store (`layout.keys`) lists a (master) key `K`,
    with a subkey `K'`, then `K'` is authorized implicitly, to sign any link
    that `K` is authorized to sign. In other words, the trust in a master key
    extends to the trust in a subkey. The inverse is not true.

  <Arguments>
    layout:
            A Layout object whose Steps are extracted and verified.

    chain_link_dict:
            A dictionary containing link metadata per functionary per step,
            e.g.:
            {
              <link name> : {
                <functionary key id> : <Metablock containing a Link or Layout
                                        object>,
                ...
              }, ...
            }

  <Exceptions>
    ThresholdVerificationError
            If any of the steps of the passed layout does not have enough
            (`step.threshold`) links signed by different authorized
            functionaries.

  <Returns>
    A chain_link_dict containing only links with valid signatures created by
    authorized functionaries.

  """
  # Create an inverse keys-subkeys dictionary, with subkey keyids as
  # dictionary keys and main keys as dictionary values. This will be
  # required below to assess main-subkey trust delegations.
  # We assume that a given subkey can only belong to one master key
  # TODO: Is this a safe assumption? Should we assert for it?
  main_keys_for_subkeys = {}
  for main_key in list(layout.keys.values()):
    for sub_keyid in main_key.get("subkeys", []):
      main_keys_for_subkeys[sub_keyid] = main_key

  # Dict for valid and authorized links of all steps of the layout
  verfied_chain_link_dict = {}

  # For each step of the layout check the signatures of corresponding links.
  # Consider only links where the signature is valid and keys are authorized,
  # and discard others.
  # Only count one of multiple links signed with different subkeys of a main
  # key towards link threshold.
  # Only proceed with final product verification if threshold requirements are
  # fulfilled.
  for step in layout.steps:
    # Dict for valid and authorized links of a given step
    verified_key_link_dict = {}
    # List of used keyids
    used_main_keyids = []

    # Do per step link threshold verification
    for link_keyid, link in chain_link_dict.get(step.name, {}).items():
      # Iterate over authorized keyids to find a key or subkey corresponding
      # to the given link and check if the link's keyid is authorized.
      # Subkeys of authorized main keys are authorized implicitly.
      for authorized_keyid in step.pubkeys:

        authorized_key = layout.keys.get(authorized_keyid)
        main_key_for_subkey = main_keys_for_subkeys.get(authorized_keyid)

        # The signing key is authorized ...
        if authorized_key and link_keyid == authorized_keyid:
          verification_key = authorized_key
          break

        # ... or the signing key is an authorized subkey ...
        if main_key_for_subkey and link_keyid == authorized_keyid:
          verification_key = main_key_for_subkey
          break

        # ... or the signing key is a subkey of an authorized key
        if (authorized_key and
            link_keyid in authorized_key.get("subkeys", {}).keys()):
          verification_key = authorized_key
          break

      else:
        LOG.info("Skipping link. Keyid '{0}' is not authorized to sign links"
            " for step '{1}'".format(link_keyid, step.name))
        continue

      # Verify signature and skip invalidly signed links
      try:
        link.verify_signature(verification_key)

      except SignatureVerificationError:
        LOG.info("Skipping link. Broken link signature with keyid '{0}'"
            " for step '{1}'".format(link_keyid, step.name))
        continue

      except KeyExpirationError as e:
        LOG.info("Skipping link. {}".format(e))
        continue

      # Warn if there are links signed by different subkeys of same main key
      if verification_key["keyid"] in used_main_keyids:
        LOG.warning("Found links signed by different subkeys of the same main"
            " key '{}' for step '{}'. Only one of them is counted towards the"
            " step threshold.".format(verification_key["keyid"], step.name))

      used_main_keyids.append(verification_key["keyid"])

      # Keep only links with valid and authorized signature
      verified_key_link_dict[link_keyid] = link

    # For each step, verify that we have enough validly signed links from
    # distinct authorized functionaries. Links signed by different subkeys of
    # the same main key are counted only once towards the threshold.
    valid_authorized_links_cnt = (len(verified_key_link_dict) -
        (len(used_main_keyids) - len(set(used_main_keyids))))
    # TODO: To guarantee that links are signed by different functionaries
    # we rely on the layout to not carry duplicate verification keys under
    # different dictionary keys, e.g. {keyid1: KEY1, keyid2: KEY1}
    # Maybe we should add such a check to the layout validation? Or here?
    if valid_authorized_links_cnt < step.threshold:
      raise ThresholdVerificationError("Step '{}' requires at least '{}' links"
          " validly signed by different authorized functionaries. Only"
          " found '{}'".format(step.name, step.threshold,
          valid_authorized_links_cnt))

    # Add all good links of this step to the dictionary of links of all steps
    verfied_chain_link_dict[step.name] = verified_key_link_dict

  # Threshold verification succeeded, return valid and authorized links for
  # further verification
  return verfied_chain_link_dict

def verify_command_alignment(command, expected_command):
  """
  <Purpose>
    Checks if a run command aligns with an expected command. The commands align
    if all of their elements are equal. If alignment fails, a warning is
    printed.

    Note:
      Command alignment is a weak guarantee. Because a functionary can easily
      alias commands.

  <Arguments>
    command:
            A command list, e.g. ["vi", "foo.py"]
    expected_command:
            A command list, e.g. ["make", "install"]

  <Exceptions>
    None.

  <Side Effects>
    Logs warning in case commands do not align.

  """
  # In what case command alignment should fail and how that failure should be
  # propagated has been thoughly discussed in:
  # https://github.com/in-toto/in-toto/issues/46 and
  # https://github.com/in-toto/in-toto/pull/47
  # We chose the simplest solution for now, i.e. Warn if they do not align.
  if command != expected_command:
    LOG.warning("Run command '{0}' differs from expected command '{1}'"
        .format(command, expected_command))


def verify_all_steps_command_alignment(layout, chain_link_dict):
  """
  <Purpose>
    Iteratively checks if all expected commands as defined in the
    Steps of a Layout align with the actual commands as recorded in the Link
    metadata.

  <Arguments>
    layout:
            A Layout object to extract the expected commands from.

    chain_link_dict:
            A dictionary containing link metadata per functionary per step,
            e.g.:
            {
              <link name> : {
                <functionary key id> : <Metablock containing a Link object>,
                ...
              }, ...
            }

  <Exceptions>
    None.

  <Side Effects>
    None.

  """
  for step in layout.steps:
    # Find the according link for this step
    expected_command = step.expected_command
    key_link_dict = chain_link_dict[step.name]

    # FIXME: I think we could do this for one link per step only
    # providing that we verify command alignment AFTER threshold equality
    for keyid, link in key_link_dict.items():
      LOG.info("Verifying command alignment for '{0}'...".format(
          in_toto.models.link.FILENAME_FORMAT.format(step_name=step.name,
              keyid=keyid)))

      command = link.signed.command
      verify_command_alignment(command, expected_command)


def verify_match_rule(rule_data, artifacts_queue, source_artifacts, links):
  """
  <Purpose>
    Filters artifacts from artifact queue using rule pattern and optional rule
    source prefix and consumes them if there is a corresponding destination
    artifact, filtered using the same rule pattern and an optional rule
    destination prefix, and source and destination artifacts have matching
    hashes.

    NOTE: The destination artifacts are extracted from the links dictionary,
    using destination name and destination type from the rule data. The source
    artifacts could also be extracted from the links dictionary, but would
    require the caller to pass source name and source type, as those are not
    encoded in the rule. However, we choose to let the caller directly pass the
    relevant artifacts.


  <Arguments>
    rule_data:
            An unpacked "MATCH" rule (see in_toto.rulelib).

    artifacts_queue:
            Not yet consumed artifacts (paths only).


    source_artifacts:
            All artifacts of the source item (including hashes).

    links:
            A dictionary containing link metadata per step or inspection, e.g.:
            {
              <link name> : <Metablock containing a link object>,
              ...
            }

  <Exceptions>
    None.

  <Side Effects>
    None.

  <Returns>
    The set of consumed artifacts (paths only).

  """
  consumed = set()

  # The rule can only consume artifacts if the destination link exists
  dest_link = links.get(rule_data["dest_name"])
  if not dest_link:
    return consumed

  # Extract destination artifacts from destination link
  dest_artifacts = getattr(dest_link.signed, rule_data["dest_type"])

  # Filter part 1 - Filter artifacts using optional source prefix, and subtract
  # prefix before filtering with rule pattern (see filter part 2) to prevent
  # globbing in the prefix.
  if rule_data["source_prefix"]:
    filtered_source_paths = []
    # Add trailing slash to source prefix if it does not exist
    normalized_source_prefix = os.path.join(
        rule_data["source_prefix"], "").replace("\\", "/")

    for artifact_path in artifacts_queue:
      if artifact_path.startswith(normalized_source_prefix):
        filtered_source_paths.append(
            artifact_path[len(normalized_source_prefix):])

  else:
    filtered_source_paths = artifacts_queue

  # Filter part 2 - glob above filtered artifact paths
  filtered_source_paths = fnmatch.filter(
      filtered_source_paths, rule_data["pattern"])

  # Iterate over filtered source paths and try to match the corresponding
  # source artifact hash with the corresponding destination artifact hash
  for path in filtered_source_paths:
    # If a source prefix was specified, we subtracted the prefix above before
    # globbing. We have to re-prepend the prefix in order to retrieve the
    # corresponding source artifact below.
    if rule_data["source_prefix"]:
      full_source_path = os.path.join(
          rule_data["source_prefix"], path).replace("\\", "/")

    else:
      full_source_path = path

    # If a destination prefix was specified, the destination artifact should
    # be queried with the full destination path, i.e. the prefix joined with
    # the globbed path.
    if rule_data["dest_prefix"]:
      full_dest_path = os.path.join(
          rule_data["dest_prefix"], path).replace("\\", "/")

    else:
      full_dest_path = path

    # Extract source artifact hash dict
    # We know the source artifact is available, it is also in the queue
    source_artifact = source_artifacts[full_source_path]

    # Don't consume source artifact w/o corresponding dest artifact (by path)
    try:
      dest_artifact = dest_artifacts[full_dest_path]
    except KeyError:
      continue

    # Don't consume source artifact w/o corresponding dest artifact (by hash)
    if source_artifact != dest_artifact:
      continue

    # Source and destination matched, consume artifact
    consumed.add(full_source_path)

  return consumed


def verify_create_rule(rule_pattern, artifacts_queue, materials, products):
  """
  <Purpose>
    Filters artifacts from artifacts queue using rule pattern and consumes them
    if they are not in the materials set but are in the products set, i.e.
    were created.

  <Arguments>
    rule_pattern:
            A "CREATE" rule pattern (see in_toto.rulelib).

    artifacts_queue:
            Not yet consumed artifacts (paths only).

    materials:
            All materials of an item (paths only).

    products:
            All products of an item (paths only).

  <Exceptions>
    None.

  <Side Effects>
    None.

  <Returns>
    The set of consumed artifacts (paths only).

  """
  # Filter queued artifacts using the rule pattern
  filtered_artifacts = fnmatch.filter(artifacts_queue, rule_pattern)

  # Consume filtered artifacts that are products but not materials
  consumed = set(filtered_artifacts) & (products - materials)

  return consumed


def verify_delete_rule(rule_pattern, artifacts_queue, materials, products):
  """
  <Purpose>
    Filters artifacts from artifacts queue using rule pattern and consumes them
    if they are in the materials set but are not in the products set, i.e.
    were deleted.

  <Arguments>
    rule_pattern:
            A "DELETE" rule pattern (see in_toto.rulelib).

    artifacts_queue:
            Not yet consumed artifacts (paths only).

    materials:
            All materials of an item (paths only).

    products:
            All products of an item (paths only).

  <Exceptions>
    None.

  <Side Effects>
    None.

  <Returns>
    The set of consumed artifacts (paths only).

  """
  # Filter queued artifacts using the rule pattern
  filtered_artifacts = fnmatch.filter(artifacts_queue, rule_pattern)

  # Consume filtered artifacts that are materials but not products
  consumed = set(filtered_artifacts) & (materials - products)

  return consumed


def verify_modify_rule(rule_pattern, artifacts_queue, materials, products):
  """
  <Purpose>
    Filters artifacts from artifacts queue using rule pattern and consumes them
    if they are in both the materials dict and in the products doct, but have
    different hashes, i.e. were modified.

  <Arguments>
    rule_pattern:
            A "MODIFY" rule pattern (see in_toto.rulelib).

    artifacts_queue:
            Not yet consumed artifacts (paths only).

    materials:
            All materials of an item (including hashes).

    products:
            All products of an item (including hashes).

  <Exceptions>
    None.

  <Side Effects>
    None.

  <Returns>
    The set of consumed artifacts (paths only).

  """
  # Filter queued artifacts using the rule pattern
  filtered_artifacts = fnmatch.filter(artifacts_queue, rule_pattern)

  # Filter filtered artifacts that are materials and products
  filtered_artifacts = set(filtered_artifacts) & \
      set(materials.keys()) & set(products.keys())

  # Consume filtered artifacts that have different hashes
  consumed = set()
  for path in filtered_artifacts:
    if materials[path] != products[path]:
      consumed.add(path)

  return consumed


def verify_allow_rule(rule_pattern, artifacts_queue):
  """
  <Purpose>
    Consumes artifacts, filtered from the artifacts queue using rule pattern.

  <Arguments>
    rule_pattern:
            An "ALLOW" rule pattern (see in_toto.rulelib).

    artifacts_queue:
            Not yet consumed artifacts (paths only).

  <Exceptions>
    None.

  <Side Effects>
    None.

  <Returns>
    The set of consumed artifacts (paths only).

  """
  # Filter queued artifacts using the rule pattern
  filtered_artifacts = fnmatch.filter(artifacts_queue, rule_pattern)

  # Consume all filtered artifacts
  return set(filtered_artifacts)


def verify_disallow_rule(rule_pattern, artifacts_queue):
  """
  <Purpose>
    Raises RuleVerificationError if rule pattern applies to any artifacts in
    the queue.

    NOTE: Each set of rules should have a terminal DISALLOW rule to make
    overall verification fail in case preceding rules did not consume all
    artifacts as intended.

  <Arguments>
    rule_pattern:
            A "DISALLOW" rule pattern (see in_toto.rulelib).

    artifacts_queue:
            Not yet consumed artifacts (paths only).

  <Exceptions>
    RuleVerificationError
        if the rule pattern filters artifacts in the artifact queue.

  <Side Effects>
    None.

  <Returns>
    None.

  """
  filtered_artifacts = fnmatch.filter(artifacts_queue, rule_pattern)

  if len(filtered_artifacts):
    raise RuleVerificationError("'DISALLOW {}' matched the following "
      "artifacts: {}\n{}".format(rule_pattern, filtered_artifacts,
      _get_artifact_rule_traceback()))


def verify_require_rule(filename, artifacts_queue):
  """
  <Purpose>
    Raises RuleVerificationError if the filename provided does not exist in the
    artifacts_queue

  <Arguments>
    filename:
            A single filename (see issues #193 and #152). We will ignore the
            artifact rule pattern because it's ambiguous and instead treat it
            as a literal file name.

    artifacts_queue:
            Not yet consumed artifacts (paths only).

  <Exceptions>
    RuleVerificationError:
      if the filename is not present in the artifacts queue

  <Side Effects>
    None.

  <Returns>
    None.

  """
  if filename not in artifacts_queue:
    raise RuleVerificationError("'REQUIRE {filename}' did not find {filename} "
        "in: {queue}\n{traceback}".format(filename=filename,
        queue=artifacts_queue, traceback=_get_artifact_rule_traceback()))


def _get_artifact_rule_traceback():
  """Build and return string form global `RULE_TRACE` which may be used as
  error message for RuleVerificationError.

  """
  traceback_str = "Full trace for 'expected_{0}' of item '{1}':\n".format(
      RULE_TRACE["source_type"], RULE_TRACE["source_name"])

  # Show all materials and products available in the beginning and
  # label the one that is used to generate a queue.
  for source_type in ["materials", "products"]:
    traceback_str += "Available {}{}:\n{}\n".format(
        source_type,
        [" (used for queue)", ""][RULE_TRACE["source_type"] != source_type],
        RULE_TRACE[source_type])

  for trace_entry in RULE_TRACE["trace"]:
    traceback_str += "Queue after '{0}':\n".format(
        " ".join(trace_entry["rule"]))
    traceback_str += "{}\n".format(trace_entry["queue"])

  return traceback_str


def verify_item_rules(source_name, source_type, rules, links):
  """
  <Purpose>
    Apply all passed material or product rules (see source_type) of a given
    step or inspection (see source_name), to enforce and authorize the
    corresponding artifacts and to guarantee that artifacts are linked together
    across steps of the supply chain.

    The mode of operation is similar to that of a firewall:
    In the beginning all materials or products of the step or inspection are
    placed into an artifact queue. The rules are then applied sequentially,
    consuming artifacts in the queue, i.e. removing them from the queue upon
    successful application.

    The consumption of artifacts by itself has no effects on the verification.
    Only through a subsequent "DISALLOW" rule, that finds unconsumed artifacts,
    is an exception raised. Similarly does the "REQUIRE" rule raise exception,
    if it does not find the artifact it requires, because it has falsely been
    consumed or was not there from the beginning.



  <Arguments>
    source_name:
            The name of the item (step or inspection) being verified.

    source_type:
            One of "materials" or "products" depending on whether the rules are
            taken from the "expected_materials" or "expected_products" field of
            the item being verified.

    rules:
            The list of rules (material or product rules) for the item being
            verified.

    links:
            A dictionary containing link metadata per step or inspection, e.g.:
            {
              <link name> : <Metablock containing a link>,
              ...
            }

  <Exceptions>
    FormatError
        if source_type is not "materials" or "products", or
        if a rule in the passed list of rules does not conform with any rule
        format.

    RuleVerificationError
        if a DISALLOW rule matches disallowed artifacts, or
        if a REQUIRE rule does not find a required artifact.

  <Side Effects>
    Clears and populates the global RULE_TRACE data structure.

  """
  if source_type not in ["materials", "products"]:
    raise securesystemslib.exceptions.FormatError(
        "Argument 'source_type' of function 'verify_item_rules' has to be "
        "one of 'materials' or 'products'. Got: '{}'".format(source_type))

  # Create shortcuts to item's materials and products (including hashes),
  # required to verify "modify" and "match" rules.
  materials_dict = links[source_name].signed.materials
  products_dict = links[source_name].signed.products

  # All other rules only require materials or products paths (without hashes)
  materials_paths = set(materials_dict.keys())
  products_paths = set(products_dict.keys())

  # Depending on the source type we create the artifact queue from the item's
  # materials or products and use it to keep track of (not) consumed artifacts.
  # The queue also only contains aritfact keys (without hashes)
  artifacts = getattr(links[source_name].signed, source_type)
  artifacts_queue = set(artifacts.keys())

  # Reset and re-populate rule traceback info dict for a rich error message
  RULE_TRACE.clear()
  RULE_TRACE["source_name"] = source_name
  RULE_TRACE["source_type"] = source_type
  RULE_TRACE["materials"] = list(materials_dict)
  RULE_TRACE["products"] = list(products_dict)
  RULE_TRACE["trace"] = []

  # Process rules and remove consumed items from queue in each iteration
  for rule in rules:
    LOG.info("Verifying '{}'...".format(" ".join(rule)))

    # Parse the rule and create two shortcuts to contained rule data
    rule_data = in_toto.rulelib.unpack_rule(rule)
    _type = rule_data["rule_type"]
    _pattern = rule_data["pattern"]


    # Initialize empty consumed set as fallback for rules that do not consume
    # artifacts. All rules except "disallow" and "require" consume artifacts.
    consumed = set()

    if _type == "match":
      consumed = verify_match_rule(
          rule_data, artifacts_queue, artifacts, links)

    elif _type == "create":
      consumed = verify_create_rule(
          _pattern, artifacts_queue, materials_paths, products_paths)

    elif _type == "delete":
      consumed = verify_delete_rule(
          _pattern, artifacts_queue, materials_paths, products_paths)

    elif _type == "modify":
      consumed = verify_modify_rule(
          _pattern, artifacts_queue, materials_dict, products_dict)

    elif _type == "allow":
      consumed = verify_allow_rule(_pattern, artifacts_queue)

    # It's up to the "disallow" and "require" rule to raise an error if
    # artifacts were not consumed as intended
    elif _type == "disallow":
      verify_disallow_rule(_pattern, artifacts_queue)

    elif _type == "require":
      verify_require_rule(_pattern, artifacts_queue)

    else: # pragma: no cover (unreachable)
      raise securesystemslib.exceptions.FormatError(
          "Invaldid rule type '{}'.".format(_type))

    artifacts_queue -= consumed

    # Append rule and copy of queue to global info for a rich error message
    RULE_TRACE["trace"].append({
          "rule": rule,
          "queue": list(artifacts_queue)
        })


def verify_all_item_rules(items, links):
  """
  <Purpose>
    Iteratively verifies artifact rules of passed items (Steps or Inspections).

  <Arguments>
    items:
            A list containing Step or Inspection objects whose material
            and product rules will be verified.

    links:
            A dictionary containing link metadata per step or inspection, e.g.:
            {
              <link name> : <Metablock containing a Link object>,
              ...
            }

  <Exceptions>
    None.

  <Side Effects>
    None.

  """
  for item in items:
    LOG.info("Verifying material rules for '{}'...".format(item.name))
    verify_item_rules(item.name, "materials", item.expected_materials, links)

    LOG.info("Verifying product rules for '{}'...".format(item.name))
    verify_item_rules(item.name, "products", item.expected_products, links)


def verify_threshold_constraints(layout, chain_link_dict):
  """
  <Purpose>
    Verifies that all links corresponding to a given step report the same
    materials and products.

    NOTE: This function does not verify if the signatures of each link
    corresponding to a step are valid or created by a different authorized
    functionary. This should be done earlier, using the function
    `verify_link_signature_thresholds`.

  <Arguments>
    layout:
            The layout whose step thresholds are being verified

    chain_link_dict:
            A dictionary containing link metadata per functionary per step,
            e.g.:
            {
              <link name> : {
                <functionary key id> : <Metablock containing a Link object>,
                ...
              }, ...
            }

  <Exceptions>
    ThresholdVerificationError
        If there are not enough (threshold) links for a steps

        If the artifacts for all links of a step are not equal


  <Side Effects>
    None.

  """

  # We are only interested in links that are related to steps defined in the
  # Layout, so iterate over layout.steps
  for step in layout.steps:
    # Skip steps that don't require multiple functionaries
    if step.threshold <= 1:
      LOG.info("Skipping threshold verification for step '{0}' with"
          " threshold '{1}'...".format(step.name, step.threshold))
      continue

    LOG.info("Verifying threshold for step '{0}' with"
        " threshold '{1}'...".format(step.name, step.threshold))
    # Extract the key_link_dict for this step from the passed chain_link_dict
    key_link_dict = chain_link_dict[step.name]

    # Check if we have at least <threshold> links for this step
    # NOTE: This is already done in `verify_link_signature_thresholds`,
    # Should we remove the check?
    if len(key_link_dict) < step.threshold:
      raise ThresholdVerificationError("Step '{0}' not performed"
          " by enough functionaries!".format(step.name))

    # Take a reference link (e.g. the first in the step_link_dict)
    reference_keyid = list(key_link_dict.keys())[0]
    reference_link = key_link_dict[reference_keyid]

    # Iterate over all links to compare their properties with a reference_link
    for keyid, link in key_link_dict.items():
      # TODO: Do we only care for artifacts, or do we want to
      # assert equality of other properties as well?
      if (reference_link.signed.materials != link.signed.materials or
          reference_link.signed.products != link.signed.products):
        raise ThresholdVerificationError("Links '{0}' and '{1}' have different"
            " artifacts!".format(
                in_toto.models.link.FILENAME_FORMAT.format(
                    step_name=step.name, keyid=reference_keyid),
                in_toto.models.link.FILENAME_FORMAT.format(
                    step_name=step.name, keyid=keyid)))


def reduce_chain_links(chain_link_dict):
  """
  <Purpose>
    Iterates through the passed chain_link_dict and builds a dict with
    step-name as keys and link objects as values.
    We already check if the links of different functionaries are
    identical.

  <Arguments>
    layout:
            The layout specified by the project owner against which the
            threshold will be verified.

    chain_link_dict:
            A dictionary containing link metadata per functionary per step,
            e.g.:
            {
              <link name> : {
                <functionary key id> : <Metablock containing a Link object>,
                ...
              }, ...
            }

  <Exceptions>
    None.

  <Side Effects>
    None.

  <Returns>
    A dictionary containing one Link metadata object per step only if
    the link artifacts of all link objects are identical for a step.

  """

  reduced_chain_link_dict = {}

  for step_name, key_link_dict in chain_link_dict.items():
    # Extract the key_link_dict for this step from the passed chain_link_dict
    # take one exemplary link (e.g. the first in the step_link_dict)
    # form the reduced_chain_link_dict to return
    reduced_chain_link_dict[step_name] = list(key_link_dict.values())[0]

  return reduced_chain_link_dict


def verify_sublayouts(layout, chain_link_dict, superlayout_link_dir_path):
  """
  <Purpose>
    Checks if any step has been delegated by the functionary, recurses into
    the delegation and replaces the layout object in the chain_link_dict
    by an equivalent link object.

  <Arguments>
    layout:
            The layout specified by the project owner.

    chain_link_dict:
            A dictionary containing link metadata per functionary per step,
            e.g.:
            {
              <link name> : {
                <functionary key id> : <Metablock containing a Link or Layout
                                          object>,
                ...
              }, ...
            }

    superlayout_link_dir_path:
            A path to a directory, where links of the superlayout are loaded
            from. Links of the sublayout are expected to be in a subdirectory
            relative to this path, with a name in the format
            in_toto.models.layout.SUBLAYOUT_LINK_DIR_FORMAT.

  <Exceptions>
    raises an Exception if verification of the delegated step fails.

  <Side Effects>
    None.

  <Returns>
    The passed dictionary containing link metadata per functionary per step,
    with layouts replaced with summary links.
    e.g.:
    {
      <link name> : {
        <functionary key id> : <Metablock containing a Link object>,
        ...
      }, ...
    }

  """
  for step_name, key_link_dict in chain_link_dict.items():

    for keyid, link in key_link_dict.items():

      if link.type_ == "layout":
        LOG.info("Verifying sublayout {}...".format(step_name))
        layout_key_dict = {}

        # Retrieve the entire key object for the keyid
        # corresponding to the link
        layout_key_dict = {keyid: layout.keys.get(keyid)}

        # Sublayout links are expected to be in a directory with the following
        # name relative the the current link directory path, i.e. if there
        # are multiple levels of sublayout nesting, the links are expected to
        # be nested accordingly
        sub_link_dir = in_toto.models.layout.SUBLAYOUT_LINK_DIR_FORMAT.format(
            name=step_name, keyid=keyid)

        sublayout_link_dir_path = os.path.join(
            superlayout_link_dir_path, sub_link_dir)

        # Make a recursive call to in_toto_verify with the
        # layout and the extracted key object
        summary_link = in_toto_verify(link, layout_key_dict,
            link_dir_path=sublayout_link_dir_path, step_name=step_name)

        # Replace the layout object in the passed chain_link_dict
        # with the link file returned by in-toto-verify
        key_link_dict[keyid] = summary_link

  return chain_link_dict


def get_summary_link(layout, reduced_chain_link_dict, name):
  """
  <Purpose>
    Merges the materials of the first step (as mentioned in the layout)
    and the products of the last step and returns a new link.
    This link reports the materials and products and summarizes the
    overall software supply chain.
    NOTE: The assumption is that the steps mentioned in the layout are
    to be performed sequentially. So, the first step mentioned in the
    layout denotes what comes into the supply chain and the last step
    denotes what goes out.

  <Arguments>
    layout:
            The layout specified by the project owner.

    reduced_chain_link_dict:
            A dictionary containing link metadata per step,
            e.g.:
            {
              <link name> : <Metablock containing a Link object>,
              ...
            }
    name:
            The name that the summary link will be associated with.

  <Exceptions>
    None.

  <Side Effects>
    None.

  <Returns>
    A Metablock object containing a Link which summarizes the materials and
    products of the overall software supply chain.

  """
  # Create empty link object
  summary_link = in_toto.models.link.Link()

  # Take first and last link in the order the corresponding
  # steps appear in the layout, if there are any.
  if len(layout.steps) > 0:
    first_step_link = reduced_chain_link_dict[layout.steps[0].name]
    last_step_link = reduced_chain_link_dict[layout.steps[-1].name]

    summary_link.materials = first_step_link.signed.materials
    summary_link.name = name

    summary_link.products = last_step_link.signed.products
    summary_link.byproducts = last_step_link.signed.byproducts
    summary_link.command = last_step_link.signed.command

  return Metablock(signed=summary_link)


def in_toto_verify(layout, layout_key_dict, link_dir_path=".",
    substitution_parameters=None, step_name="",
    persist_inspection_links=True):
  """Performs complete in-toto supply chain verification for a final product.

  The verification procedure consists of the following activities, performed in
  the given order:

  1.  Verify layout signatures
  2.  Verify layout expiration date
  3.  Substitute placeholders in the layout
  4.  Load link metadata
  5.  Verify link signatures with keys in layout
  6.  Recurse into sublayout verification
  7.  Soft-verify alignment of reported and expected commands. Note:
      verification does not fail if commands don't match, instead a warning is
      logged as per the in-toto specification --
      https://github.com/in-toto/docs/blob/master/in-toto-spec.md#431-steps
  8.  Verify threshold artifact constraints
  9.  Process step product and material rules
  10. Execute inspection commands and generate inspection links
  11. Process inspection product and material rules

  Arguments:
    layout: A Metablock object that contains a Layout object to be verified.

    layout_key_dict: A public key dictionary. The verification routine requires
        at least one key, and a valid signature on the layout for each key.

    link_dir_path (optional): A directory path to link metadata files. The
        expected filename format for link metadata files is
        ``STEP-NAME.KEYID-PREFIX.link``. Link metadata files for a sublayout
        are loaded from a subdirectory relative to the link_dir_path of the
        superlayout. The expected directory name format is
        ``SUBLAYOUT-STEP-NAME.KEYID-PREFIX``.

    substitution_parameters (optional): A dictionary with substitution values
        for artifact rules (steps and inspections), the expected command
        attribute (steps), and the run attribute (inspections) in the layout.

    step_name (optional): A name assigned to the returned link. This is mostly
        useful during recursive sublayout verification.

    persist_inspection_links (optional): A boolean that determines whether or
        not link metadata files for inspection are written to cwd.

  Raises:
    securesystemslib.exceptions.FormatError: Passed parameters are malformed.

    SignatureVerificationError: No layout verification key is passed, or any of
        the passed keys fails to verify a signature.

    LayoutExpiredError: The layout is expired.

    LinkNotFoundError: Fewer than threshold link metadata files can be found
        for a step of the layout.

    ThresholdVerificationError: Fewer than threshold links, validly signed by
        different authorized functionaries, who agree on the recorded materials
        and products, can be found for a step of the layout. (Links with
        invalid signatures or signatures by unauthorized functionaries are
        ignored.)

    RuleVerificationError: A DISALLOW rule matches disallowed artifacts, or
        a REQUIRE rule does not find a required artifact.

    BadReturnValueError: An inspection command returns a non-zero value.

  Side Effects:
    Reads link metadata files from disk.
    Runs inspection commands in subprocess.

  Returns:
    A Metablock object that contains a Link object, which summarizes the
    materials and products of the overall software supply chain. This is mostly
    useful during recursive sublayout verification.

  """
  LOG.info("Verifying layout signatures...")
  verify_layout_signatures(layout, layout_key_dict)

  # For the rest of the verification we only care about the layout payload
  # (Layout) that carries all the information and not about the layout
  # container (Metablock) that also carries the signatures
  layout = layout.signed

  LOG.info("Verifying layout expiration...")
  verify_layout_expiration(layout)

  # If there are parameters sent to the translation layer, substitute them
  if substitution_parameters is not None:
    LOG.info('Performing parameter substitution...')
    substitute_parameters(layout, substitution_parameters)

  LOG.info("Reading link metadata files...")
  chain_link_dict = load_links_for_layout(layout, link_dir_path)

  LOG.info("Verifying link metadata signatures...")
  chain_link_dict = verify_link_signature_thresholds(layout, chain_link_dict)

  LOG.info("Verifying sublayouts...")
  chain_link_dict = verify_sublayouts(layout, chain_link_dict, link_dir_path)

  LOG.info("Verifying alignment of reported commands...")
  verify_all_steps_command_alignment(layout, chain_link_dict)

  LOG.info("Verifying threshold constraints...")
  verify_threshold_constraints(layout, chain_link_dict)
  reduced_chain_link_dict = reduce_chain_links(chain_link_dict)

  # NOTE: Processing step rules before executing inspections guarantees that
  # inspection commands don't run on compromised target files, however, this
  # also precludes step match rules from referencing inspection artifacts.
  LOG.info("Verifying Step rules...")
  verify_all_item_rules(layout.steps, reduced_chain_link_dict)

  LOG.info("Executing Inspection commands...")
  inspection_link_dict = run_all_inspections(layout, persist_inspection_links)

  LOG.info("Verifying Inspection rules...")
  # Artifact rules for inspections can reference links that correspond to
  # Steps or Inspections, hence the concatenation of both collections of links
  combined_links = reduced_chain_link_dict.copy()
  combined_links.update(inspection_link_dict)
  verify_all_item_rules(layout.inspect, combined_links)

  # We made it this far without exception that means, verification passed
  LOG.info("The software product passed all verification.")

  # Return a link file which summarizes the entire software supply chain
  # This is mostly relevant if the currently verified supply chain is embedded
  # in another supply chain
  return get_summary_link(layout, reduced_chain_link_dict, step_name)



#!/usr/bin/env python

# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Program Name>
  layout.py

<Author>
  Lukas Puehringer <lukas.puehringer@nyu.edu>
  Santiago Torres <santiago@nyu.edu>

<Started>
  Sep 23, 2016

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides classes related to the definition of a software supply chain.

<Classes>
  Layout:
      represents the metadata file that defines a software supply chain through
      steps and inspections

  Step:
      represents one step of the software supply chain, performed by one or
      many functionaries, who are identified by a key also stored to the layout

  Inspection:
      represents a hook that is run at verification
"""
import attr
import shlex
import json

from datetime import datetime
from dateutil.relativedelta import relativedelta
from dateutil.parser import parse

from in_toto.models.common import Signable, ValidationMixin
import in_toto.rulelib
import in_toto.exceptions

import securesystemslib.exceptions
import securesystemslib.formats
import securesystemslib.schema
import securesystemslib.interface
import securesystemslib.gpg.functions


# Link metadata for sublayouts are expected to be found in a subdirectory
# with the following name, relative to the verification directory
SUBLAYOUT_LINK_DIR_FORMAT = "{name}.{keyid:.8}"



@attr.s(repr=False, init=False)
class Layout(Signable):
  """A definition for a software supply chain.

  A layout lists the sequence of steps of the software supply chain in the
  order they are expected to be performed, the functionaries authorized and
  required to perform them, and inspections to be performed by the client upon
  final product verification.

  A Layout object is usually contained in a generic Metablock object for
  signing, serialization and I/O capabilities.

  Attributes:
    steps: A list of Step objects.

    inspect: A list of Inspection objects.

    keys: A dictionary of functionary public keys, with keyids as dict keys and
        keys as values.

    expires: The layout expiration.

    readme: A human readable description of the software supply chain.

  """
  _type = attr.ib()
  steps = attr.ib()
  inspect = attr.ib()
  keys = attr.ib()
  expires = attr.ib()
  readme = attr.ib()


  def __init__(self, **kwargs):
    super(Layout, self).__init__()
    self._type = "layout"
    self.steps = kwargs.get("steps", [])
    self.inspect = kwargs.get("inspect", [])
    self.keys = kwargs.get("keys", {})
    self.readme = kwargs.get("readme", "")

    # Assign a default expiration (one month) if not expiration date is passed
    # TODO: Is one month a sensible default? In any case, we need a valid
    # expiration date in order for the layout object to validate.
    # (see self._validate_expires)
    self.expires = kwargs.get("expires")
    if not self.expires:
      self.set_relative_expiration(months=1)

    # TODO: Should we validate in the constructor or allow the user to create
    # an invalid layout and call validate explicitly?
    self.validate()


  @property
  def type_(self):
    """The string "layout" to indentify the in-toto metadata type."""
    # NOTE: We expose the type_ property in the API documentation instead of
    # _type to protect it against modification.
    # NOTE: Trailing underscore is used by convention (pep8) to avoid conflict
    # with Python's type keyword.
    return self._type


  @staticmethod
  def read(data):
    """Creates a Layout object from its dictionary representation.

    Arguments:
      data: A dictionary with layout metadata fields.

    Raises:
      securesystemslib.exceptions.FormatError: Passed data is invalid.

    Returns:
      The created Layout object.

    """
    steps = []

    for step_data in data.get("steps"):
      steps.append(Step.read(step_data))
    data["steps"] = steps

    inspections = []
    for inspect_data in data.get("inspect"):
      inspections.append(Inspection.read(inspect_data))
    data["inspect"] = inspections

    return Layout(**data)


  def set_relative_expiration(self, days=0, months=0, years=0):
    """Sets layout expiration relative to today.

    If not argument is passed the set exipration date is now.

    Arguments:
      days (optional): Days from today.
      months (optional): Months from today.
      years (optional): Years from today.

    Raises:
      securesystemslib.exceptions.FormatError: Arguments are not ints.

    """
    securesystemslib.schema.Integer().check_match(days)
    securesystemslib.schema.Integer().check_match(months)
    securesystemslib.schema.Integer().check_match(years)

    self.expires = (datetime.today() + relativedelta(
        days=days, months=months, years=years)).strftime("%Y-%m-%dT%H:%M:%SZ")


  def get_step_name_list(self):
    """Returns ordered list of step names as they appear in the layout.

    Returns:
      A list of step names.

    """
    step_names = []
    for step in self.steps:
      step_names.append(step.name)

    return step_names


  def get_step_by_name(self, step_name):
    """Returns step identified by step_name from the layout.

    NOTE: Returns the first step identified only, which should be the only step
    for a given name of a valid layout.

    Arguments:
      step_name: A step name.

    Raises:
      securesystemslib.exceptions.FormatError: Argument is not a string.

    Returns:
      A Step object.

    """
    securesystemslib.schema.AnyString().check_match(step_name)

    for step in self.steps: # pragma: no branch
      if step.name == step_name:
        return step


  def remove_step_by_name(self, step_name):
    """Removes steps identified by step_name from the layout.

    NOTE: Removes all steps identified, which should be only one step for a
    given name of a valid layout.

    Arguments:
      step_name: A step name.

    Raises:
      securesystemslib.exceptions.FormatError: Argument is not a string.

    """
    securesystemslib.schema.AnyString().check_match(step_name)

    for step in self.steps:
      if step.name == step_name:
        self.steps.remove(step)


  def get_inspection_name_list(self):
    """Returns ordered list of inspection names as they appear in the layout.

    Returns:
      A list of inspection names.

    """
    inspection_names = []
    for inspection in self.inspect:
      inspection_names.append(inspection.name)

    return inspection_names


  def get_inspection_by_name(self, inspection_name):
    """Returns inspection identified by inspection_names from the layout.

    NOTE: Returns the first inspection identified only, which should be the
    only inspection for a given name of a valid layout.

    Arguments:
      inspection_name: An inspection name.

    Raises:
      securesystemslib.exceptions.FormatError: Argument is not a string.

    Returns:
      An Inspection object.

    """
    securesystemslib.schema.AnyString().check_match(inspection_name)

    for inspection in self.inspect: # pragma: no branch
      if inspection.name == inspection_name:
        return inspection


  def remove_inspection_by_name(self, inspection_name):
    """Removes inspections identified by inspection_name from the layout.

    NOTE: Removes all inspections identified, which should be only one
    inspection for a given name of a valid layout.

    Arguments:
      inspection_name: An inspection name.

    Raises:
      securesystemslib.exceptions.FormatError: Argument is not a string.

    """
    securesystemslib.schema.AnyString().check_match(inspection_name)

    for inspection in self.inspect:
      if inspection.name == inspection_name:
        self.inspect.remove(inspection)


  def get_functionary_key_id_list(self):
    """Returns list of functionary keyids from the layout.

    Returns:
      A list of keyids.

    """
    return list(self.keys.keys())


  def add_functionary_key(self, key):
    """Adds key as functionary key to layout.

    Arguments:
      key: A public key. Format is securesystemslib.formats.ANY_PUBKEY_SCHEMA.

    Raises:
      securesystemslib.exceptions.FormatError: Argument is malformed.

    Returns:
      The added key.

    """
    securesystemslib.formats.ANY_PUBKEY_SCHEMA.check_match(key)
    keyid = key["keyid"]
    self.keys[keyid] = key
    return key


  def add_functionary_key_from_path(self, key_path):
    """Loads key from disk and adds as functionary key to layout.

    Arguments:
      key_path: A path to a PEM-formatted RSA public key. Format is
          securesystemslib.formats.PATH_SCHEMA.

    Raises:
      securesystemslib.exceptions.FormatError: Argument is malformed.
      securesystemslib.exceptions.Error: Key cannot be imported.

    Returns:
      The added functionary public key.

    """
    securesystemslib.formats.PATH_SCHEMA.check_match(key_path)
    key = securesystemslib.interface.import_rsa_publickey_from_file(key_path)

    return self.add_functionary_key(key)



  def add_functionary_key_from_gpg_keyid(self, gpg_keyid, gpg_home=None):
    """Loads key from gpg keychain and adds as functionary key to layout.

    Arguments:
      gpg_keyid: A keyid used to identify a local gpg public key.
      gpg_home (optional): A path to the gpg home directory. If not set the
          default gpg home directory is used.

    Raises:
      securesystemslib.exceptions.FormatError: Arguments are malformed.
      securesystemslib.gpg.execeptions.KeyNotFoundError: Key cannot be found.

    Side Effects:
      Calls system gpg command in a subprocess.

    Returns:
      The added key.

    """
    securesystemslib.formats.KEYID_SCHEMA.check_match(gpg_keyid)
    if gpg_home: # pragma: no branch
      securesystemslib.formats.PATH_SCHEMA.check_match(gpg_home)

    key = securesystemslib.gpg.functions.export_pubkey(gpg_keyid,
        homedir=gpg_home)
    return self.add_functionary_key(key)


  def add_functionary_keys_from_paths(self, key_path_list):
    """Loads keys from disk and adds as functionary keys to layout.

    Arguments:
      key_path_list: A list of paths to PEM-formatted RSA public keys. Format
          of each path is securesystemslib.formats.PATH_SCHEMA.

    Raises:
      securesystemslib.exceptions.FormatError: Argument is malformed.
      securesystemslib.exceptions.Error: A key cannot be imported.

    Returns:
      A dictionary of the added functionary keys, with keyids as dictionary
      keys and keys as values.

    """
    securesystemslib.formats.PATHS_SCHEMA.check_match(key_path_list)
    key_dict = {}
    for key_path in key_path_list:
      key = self.add_functionary_key_from_path(key_path)
      key_dict[key["keyid"]] = key

    return key_dict


  def add_functionary_keys_from_gpg_keyids(self, gpg_keyid_list,
      gpg_home=None):
    """Loads keys from gpg keychain and adds as functionary keys to layout.

    Arguments:
      gpg_keyid_list: A list of keyids used to identify local gpg public keys.
      gpg_home (optional): A path to the gpg home directory. If not set the
          default gpg home directory is used.

    Raises:
      securesystemslib.exceptions.FormatError: Arguments are malformed.
      securesystemslib.gpg.execeptions.KeyNotFoundError: A key cannot be found.

    Side Effects:
      Calls system gpg command in a subprocess.

    Returns:
      A dictionary of the added functionary keys, with keyids as dictionary
      keys and keys as values.

    """
    securesystemslib.formats.KEYIDS_SCHEMA.check_match(gpg_keyid_list)
    key_dict = {}
    for gpg_keyid in gpg_keyid_list:
      key = self.add_functionary_key_from_gpg_keyid(gpg_keyid, gpg_home)
      key_dict[key["keyid"]] = key

    return key_dict


  def _validate_type(self):
    """Private method to check that the type string is set to layout."""
    if self._type != "layout":
      raise securesystemslib.exceptions.FormatError(
          "Invalid _type value for layout (Should be 'layout')")


  def _validate_expires(self):
    """Private method to verify if the expiration field has the right format
    and can be parsed."""
    try:
      # We do both 'parse' and 'check_match' because the format check does not
      # detect bogus dates (e.g. Jan 35th) and parse can do more formats.
      parse(self.expires)
      securesystemslib.formats.ISO8601_DATETIME_SCHEMA.check_match(
          self.expires)
    except Exception as e:
      raise securesystemslib.exceptions.FormatError(
          "Malformed date string in layout. Exception: {}".format(e))


  def _validate_readme(self):
    """Private method to check that the readme field is a string."""
    if not isinstance(self.readme, str):
      raise securesystemslib.exceptions.FormatError(
          "Invalid readme '{}', value must be a string."
          .format(self.readme))


  def _validate_keys(self):
    """Private method to ensure that the keys contained are right."""
    securesystemslib.formats.ANY_PUBKEY_DICT_SCHEMA.check_match(self.keys)


  def _validate_steps_and_inspections(self):
    """Private method to verify that the list of steps and inspections are
    correctly formed."""
    names_seen = set()
    if not isinstance(self.steps, list):
      raise securesystemslib.exceptions.FormatError(
          "The steps field should be a list!")

    for step in self.steps:
      if not isinstance(step, Step):
        raise securesystemslib.exceptions.FormatError(
            "The steps list should only contain steps!")

      step.validate()

      if step.name in names_seen:
        raise securesystemslib.exceptions.FormatError(
            "There is already a step with name '{}'. Step names must be"
            " unique within a layout.".format(step.name))
      names_seen.add(step.name)

    if not isinstance(self.inspect, list):
      raise securesystemslib.exceptions.FormatError(
          "The inspect field should a be a list!")

    for inspection in self.inspect:
      if not isinstance(inspection, Inspection):
        raise securesystemslib.exceptions.FormatError(
            "The inspect list should only contain inspections!")

      inspection.validate()

      if inspection.name in names_seen:
        raise securesystemslib.exceptions.FormatError(
            "There is already an inspection with name '{}'. Inspection names"
            " must be unique within a layout.".format(inspection.name))
      names_seen.add(inspection.name)



@attr.s(repr=False, init=False)
class SupplyChainItem(ValidationMixin):
  """Common attributes and methods for supply chain steps and inspections.

  Attributes:
    name: A unique named used to associate related link metadata.

    expected_materials: A list of rules to encode expectations about used
        artifacts (see ``rulelib`` for formats).

    expected_products:  A list of rules to encode expectations about produced
        artifacts (see ``rulelib`` for formats).

  """
  name = attr.ib()
  expected_materials = attr.ib()
  expected_products = attr.ib()


  def __init__(self, **kwargs):
    super(SupplyChainItem, self).__init__()
    self.name = kwargs.get("name")
    self.expected_materials = kwargs.get("expected_materials", [])
    self.expected_products = kwargs.get("expected_products", [])


  def __repr__(self):
    """Returns an indented JSON string of the metadata object. """
    return json.dumps(attr.asdict(self),
        indent=1, separators=(",", ": "), sort_keys=True)


  def add_material_rule_from_string(self, rule_string):
    """Parse artifact rule string as list and add to expected_materials.

    Arguments:
      rule_string: An artifact rule string (see ``rulelib`` for formats).

    Raises:
      securesystemslib.exceptions.FormatError: Argument is malformed.

    """
    securesystemslib.schema.AnyString().check_match(rule_string)
    rule_list = shlex.split(rule_string)

    # Raises format error if the parsed rule_string is not a valid rule
    in_toto.rulelib.unpack_rule(rule_list)

    self.expected_materials.append(rule_list)


  def add_product_rule_from_string(self, rule_string):
    """Parse artifact rule string as list and add to expected_products.

    Arguments:
      rule_string: An artifact rule string (see ``rulelib`` for formats).

    Raises:
      securesystemslib.exceptions.FormatError: Argument is malformed.

    """
    securesystemslib.schema.AnyString().check_match(rule_string)
    rule_list = shlex.split(rule_string)

    # Raises format error if the parsed rule_string is not a valid rule
    in_toto.rulelib.unpack_rule(rule_list)

    self.expected_products.append(rule_list)


  def _validate_expected_materials(self):
    """Private method to check that material rules are correctly formed."""
    if not isinstance(self.expected_materials, list):
      raise securesystemslib.exceptions.FormatError(
          "Material rules should be a list!")

    for rule in self.expected_materials:
      in_toto.rulelib.unpack_rule(rule)


  def _validate_expected_products(self):
    """Private method to check that product rules are correctly formed."""
    if not isinstance(self.expected_products, list):
      raise securesystemslib.exceptions.FormatError(
          "Product rules should be a list!")

    for rule in self.expected_products:
      in_toto.rulelib.unpack_rule(rule)



@attr.s(repr=False, init=False)
class Step(SupplyChainItem):
  """A step of a software supply chain.

  A Step object is usually contained in a Layout object and encodes the
  expectations for a step of the software supply chain such as, who is
  authorized to perform the step, what command is executed, and which artifacts
  are used and produced. Evidence about a performed step is provided by link
  metadata.

  Attributes:
    pubkeys: A list of functionary keyids authorized to perform the step.

    threshold: A minimum number of distinct functionaries required to provide
        evidence for a step.

    expected_command: A list of command and command arguments, expected to
        perform the step.

  """
  _type = attr.ib()
  pubkeys = attr.ib()
  expected_command = attr.ib()
  threshold = attr.ib()


  def __init__(self, **kwargs):
    super(Step, self).__init__(**kwargs)
    self._type = "step"
    self.pubkeys = kwargs.get("pubkeys", [])
    self.expected_command = kwargs.get("expected_command", [])
    self.threshold = kwargs.get("threshold", 1)

    self.validate()


  @staticmethod
  def read(data):
    """Creates a Step object from its dictionary representation.

    Arguments:
      data: A dictionary with step metadata fields.

    Raises:
      securesystemslib.exceptions.FormatError: Passed data is invalid.

    Returns:
      The created Step object.

    """
    return Step(**data)


  def set_expected_command_from_string(self, command_string):
    """Parse command string as list and assign to expected_command.

    Arguments:
      command_string: A command and command arguments string.

    Raises:
      securesystemslib.exceptions.FormatError: Argument is malformed.

    """
    securesystemslib.schema.AnyString().check_match(command_string)
    self.expected_command = shlex.split(command_string)


  def _validate_type(self):
    """Private method to ensure that the type field is set to step."""
    if self._type != "step":
      raise securesystemslib.exceptions.FormatError(
          "Invalid _type value for step (Should be 'step')")


  def _validate_threshold(self):
    """Private method to check that the threshold field is set to an int."""
    if not isinstance(self.threshold, int):
      raise securesystemslib.exceptions.FormatError(
          "Invalid threshold '{}', value must be an int."
          .format(self.threshold))


  def _validate_pubkeys(self):
    """Private method to check that the pubkeys is a list of keyids."""
    if not isinstance(self.pubkeys, list):
      raise securesystemslib.exceptions.FormatError(
          "The pubkeys field should be a list!")

    for keyid in self.pubkeys:
      securesystemslib.formats.KEYID_SCHEMA.check_match(keyid)


  def _validate_expected_command(self):
    """Private method to check that the expected_command is proper."""
    if not isinstance(self.expected_command, list):
      raise securesystemslib.exceptions.FormatError(
          "The expected command field is malformed!")



@attr.s(repr=False, init=False)
class Inspection(SupplyChainItem):
  """An inspection for a software supply chain.

  An Inspection object is usually contained in a Layout object and encodes a
  command to be executed by an in-toto client during final product
  verification. Akin to steps, inspections can define artifact rules.

  Attributes:
    run: A list of command and command arguments to be executed upon final
        product verification.

  """
  _type = attr.ib()
  run = attr.ib()


  def __init__(self, **kwargs):
    super(Inspection, self).__init__(**kwargs)
    self._type = "inspection"
    self.run = kwargs.get("run", [])

    self.validate()


  @staticmethod
  def read(data):
    """Creates an Inspection object from its dictionary representation.

    Arguments:
      data: A dictionary with inspection metadata fields.

    Raises:
      securesystemslib.exceptions.FormatError: Passed data is invalid.

    Returns:
      The created Inspection object.

    """
    return Inspection(**data)


  def set_run_from_string(self, command_string):
    """Parse command string as list and assign to run attribute.

    Arguments:
      command_string: A command and command arguments string.

    Raises:
      securesystemslib.exceptions.FormatError: Argument is malformed.

    """
    securesystemslib.schema.AnyString().check_match(command_string)
    self.run = shlex.split(command_string)


  def _validate_type(self):
    """Private method to ensure that the type field is set to inspection."""
    if self._type != "inspection":
      raise securesystemslib.exceptions.FormatError(
          "The _type field must be set to 'inspection'!")


  def _validate_run(self):
    """Private method to check that the expected command is correct."""
    if not isinstance(self.run, list):
      raise securesystemslib.exceptions.FormatError(
          "The run field is malformed!")



# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Module Name>
  in_toto/rulelib.py

<Author>
  Santiago Torres-Arias <santiago@nyu.edu>
  Lukas Puehringer <lukas.puehringer@nyu.edu>

<Started>
  Nov 18, 2016

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  This module provides functions parse artifact rules and validate their
  syntax.

"""
import securesystemslib.exceptions
import securesystemslib.formats

GENERIC_RULES = {"create", "modify", "delete", "allow", "disallow", "require",}
COMPLEX_RULES = {"match",}
ALL_RULES = GENERIC_RULES | COMPLEX_RULES

def unpack_rule(rule):
  """
  Parses the rule and extracts and returns the necessary data to apply the
  rule. Can also be used to verify if a rule complies with any of the formats

  <Arguments>
    rule:
        The list of rule elements, in one of the following formats:
        MATCH <pattern> [IN <source-path-prefix>] WITH (MATERIALS|PRODUCTS)
            [IN <destination-path-prefix>] FROM <step>,
        CREATE <pattern>,
        DELETE <pattern>,
        MODIFY <pattern>,
        ALLOW <pattern>,
        DISALLOW <pattern>,
        REQUIRE <file>

  Note that REQUIRE is somewhat of a weird animal that does not use patterns
  but rather single filenames (for now).

  <Exceptions>
    raises FormatError, if the rule does not comply with any of the formats.

  <Side Effects>
    None.

  <Returns>
    A dictionary of the artifact rule data,
    if it is a generic rule the dictionary is:
    {
      "rule_type": rule[0] ("CREATE"|"MODIFY"|"DELETE"|"ALLOW"|"DISALLOW")
      "pattern" : rule[1], a path pattern
    }

    if it is a match rule, the dictionary is:
    {
      "rule_type": rule[0],  ("MATCH"),
      "pattern": rule[1], a path pattern
      "source_prefix": path or empty string
      "dest_prefix": path or empty string
      "dest_type" : destination artifact type, ("MATERIAL"|"PRODUCT")
      "dest_name": destination step/inspection name
    }

  """
  securesystemslib.formats.LIST_OF_ANY_STRING_SCHEMA.check_match(rule)

  # Create all lower rule copy to case insensitively parse out tokens whose
  # position we don't know yet
  # We keep the original rule to retain the non-token elements' case
  rule_lower = []
  for rule_elem in rule:
    rule_lower.append(rule_elem.lower())

  rule_len = len(rule)

  if rule_len < 2 or rule_lower[0] not in ALL_RULES:
    raise securesystemslib.exceptions.FormatError("Wrong rule format,"
        " rules must start with one of '{0}' and specify a 'pattern' as"
        " second element.\n"
        "Got: \n\t'{1}'".format(", ".join(ALL_RULES), rule))

  rule_type = rule_lower[0]
  pattern = rule[1]

  # Type is one of "CREATE", "MODIFY", "DELETE", "ALLOW", "DISALLOW"
  if rule_type in GENERIC_RULES:
    # pylint: disable=no-else-raise
    if rule_len != 2:
      raise securesystemslib.exceptions.FormatError("Wrong rule format,"
        " generic rules must have one of the formats:\n\t"
        "CREATE <pattern>\n\t"
        "MODIFY <pattern>\n\t"
        "DELETE <pattern>\n\t"
        "ALLOW <pattern>\n\t"
        "DISALLOW <pattern>\n"
        "REQUIRE <file>\n"
        "Got:\n\t{}".format(rule))
    else:
      return {
        "rule_type": rule_type,
        "pattern": pattern,
      }

  # Type is "MATCH"
  # NOTE: Can't reach `else` branch, if the rule is neither in GENERIC_RULES
  # nor in COMPLEX_RULES an exception is raised earlier.
  elif rule_type in COMPLEX_RULES: # pragma: no branch
    # ... IN <source-path-prefix> WITH (MATERIALS|PRODUCTS)
    # IN <destination-path-prefix> FROM <step>
    if (rule_len == 10 and rule_lower[2] == "in" and
        rule_lower[4] == "with" and rule_lower[6] == "in" and
        rule_lower[8] == "from"):
      source_prefix = rule[3]
      dest_type = rule_lower[5]
      dest_prefix = rule[7]
      dest_name = rule[9]

    # ... IN <source-path-prefix> WITH (MATERIALS|PRODUCTS) FROM <step>
    elif (rule_len == 8 and rule_lower[2] == "in" and
        rule_lower[4] == "with" and rule_lower[6] == "from"):
      source_prefix = rule[3]
      dest_type = rule_lower[5]
      dest_prefix = ""
      dest_name = rule[7]

    # ... WITH (MATERIALS|PRODUCTS) IN <destination-path-prefix> FROM <step>
    elif (rule_len == 8 and rule_lower[2] == "with" and
        rule_lower[4] == "in" and rule_lower[6] == "from"):
      source_prefix = ""
      dest_type = rule_lower[3]
      dest_prefix = rule[5]
      dest_name = rule[7]

    # ... WITH (MATERIALS|PRODUCTS) FROM <step>
    elif (rule_len == 6 and rule_lower[2] == "with" and
        rule_lower[4] == "from"):
      source_prefix = ""
      dest_type = rule_lower[3]
      dest_prefix = ""
      dest_name = rule[5]

    else:
      raise securesystemslib.exceptions.FormatError("Wrong rule format,"
          " match rules must have the format:\n\t"
          " MATCH <pattern> [IN <source-path-prefix>] WITH"
          " (MATERIALS|PRODUCTS) [IN <destination-path-prefix>] FROM <step>.\n"
          "Got: \n\t{}".format(rule))

    if dest_type not in {"materials", "products"}:
      raise securesystemslib.exceptions.FormatError("Wrong rule format,"
          " match rules must have either MATERIALS or PRODUCTS (case"
          " insensitive) as destination.\n"
          "Got: \n\t{}".format(rule))

    return {
      "rule_type": rule_type,
      "pattern": pattern,
      "source_prefix": source_prefix,
      "dest_prefix": dest_prefix,
      "dest_type" : dest_type,
      "dest_name": dest_name
    }


def pack_rule(rule_type, pattern, source_prefix=None, dest_type=None,
      dest_prefix=None, dest_name=None):
  """
  <Purpose>
    Create an artifact rule in the passed arguments and return it as list
    as it is stored in a step's or inspection's expected_material or
    expected_product field.

  <Arguments>
    rule_type:
            One of "MATCH", "CREATE", "DELETE", MODIFY, ALLOW, DISALLOW,
            REQUIRE (case insensitive).

    pattern:
            A glob pattern to match artifact paths.

    source_prefix: (only for "MATCH" rules)
            A prefix for 'pattern' to match artifacts reported by the link
            corresponding to the step that contains the rule.

    dest_type: (only for "MATCH" rules)
            One of "MATERIALS" or "PRODUCTS" (case insensitive) to specify
            if materials or products of the link corresponding to the step
            identified by 'dest_name' should be matched.

    dest_prefix: (only for "MATCH" rules)
            A prefix for 'pattern' to match artifacts reported by the link
            corresponding to the step identified by 'dest_name'.

    dest_name: (only for "MATCH" rules)
            The name of the step whose corresponding link is used to match
            artifacts.


  <Exceptions>
    securesystemslib.exceptions.FormatError
            if any of the arguments is malformed.


  <Returns>
    One of the following rule formats in as a list
      MATCH <pattern> [IN <source-path-prefix>] WITH (MATERIALS|PRODUCTS)
          [IN <destination-path-prefix>] FROM <step>,
      CREATE <pattern>,
      DELETE <pattern>,
      MODIFY <pattern>,
      ALLOW <pattern>,
      DISALLOW <pattern>,
      REQUIRE <file>

    Note that REQUIRE is somewhat of a weird animal that does not use patterns
    but rather single filenames (for now).
  """
  securesystemslib.formats.ANY_STRING_SCHEMA.check_match(rule_type)
  securesystemslib.formats.ANY_STRING_SCHEMA.check_match(pattern)

  if rule_type.lower() not in ALL_RULES:
    raise securesystemslib.exceptions.FormatError("'{0}' is not a valid "
        "'type'.  Rule type must be one of:  {1} (case insensitive)."
        .format(rule_type, ", ".join(ALL_RULES)))

  if rule_type.upper() == "MATCH":
    if (not securesystemslib.formats.ANY_STRING_SCHEMA.matches(dest_type) or
        not (dest_type.lower() == "materials" or
        dest_type.lower() == "products")):
      raise securesystemslib.exceptions.FormatError("'{}' is not a valid"
          " 'dest_type'. Rules of type 'MATCH' require a destination type of"
          " either 'MATERIALS' or 'PRODUCTS' (case insensitive)."
          .format(dest_type))

    if not (securesystemslib.formats.ANY_STRING_SCHEMA.matches(dest_name) and
        dest_name):
      raise securesystemslib.exceptions.FormatError("'{}' is not a valid"
          " 'dest_name'. Rules of type 'MATCH' require a step name as a"
          " destination name.".format(dest_name))

    # Construct rule
    rule = ["MATCH", pattern]

    if source_prefix:
      securesystemslib.formats.ANY_STRING_SCHEMA.check_match(source_prefix)
      rule += ["IN", source_prefix]

    rule += ["WITH", dest_type.upper()]

    if dest_prefix:
      securesystemslib.formats.ANY_STRING_SCHEMA.check_match(dest_prefix)
      rule += ["IN", dest_prefix]

    rule += ["FROM", dest_name]

  else:
    rule = [rule_type.upper(), pattern]

  return rule


def pack_rule_data(rule_data):
  """Shortcut for 'pack_rule' to pack a rule based on a rule_data dictionary
  as returned by 'unpack_rule'. """
  return pack_rule(**rule_data)


def pack_create_rule(pattern):
  """Shortcut for 'pack_rule' to pack a CREATE rule. """
  return pack_rule("CREATE", pattern)


def pack_delete_rule(pattern):
  """Shortcut for 'pack_rule' to pack a DELETE rule. """
  return pack_rule("DELETE", pattern)


def pack_modify_rule(pattern):
  """Shortcut for 'pack_rule' to pack a MODIFY rule. """
  return pack_rule("MODIFY", pattern)


def pack_allow_rule(pattern):
  """Shortcut for 'pack_rule' to pack an ALLOW rule. """
  return pack_rule("ALLOW", pattern)


def pack_disallow_rule(pattern):
  """Shortcut for 'pack_rule' to pack a DISALLOW rule. """
  return pack_rule("DISALLOW", pattern)

def pack_require_rule(filename):
  """
  Shortcut for 'pack_rule' to pack a REQUIRE rule:
  note that REQUIRE is somewhat of a weird animal that does not use patterns
  but rather single filenames (for now).
  """
  return pack_rule("REQUIRE", filename)



# Copyright New York University and the in-toto contributors
# SPDX-License-Identifier: Apache-2.0

"""
<Program Name>
  metadata.py

<Author>
  Lukas Puehringer <lukas.puehringer@nyu.edu>
  Santiago Torres <santiago@nyu.edu>

<Started>
  Oct 23, 2017

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides a container class `Metablock` for signed metadata and
  functions for signing, signature verification, de-serialization and
  serialization from and to JSON.

"""

import attr
import json

import securesystemslib.keys
import securesystemslib.formats
import securesystemslib.exceptions
import securesystemslib.gpg.functions

from in_toto.models.common import ValidationMixin
from in_toto.models.link import Link
from in_toto.models.layout import Layout
from in_toto.exceptions import SignatureVerificationError

@attr.s(repr=False, init=False)
class Metablock(ValidationMixin):
  """A container for signed in-toto metadata.

  Provides methods for metadata JSON (de-)serialization, reading from and
  writing to disk, creating and verifying signatures, and self-validation.

  Attributes:
    signed: A subclass of Signable which has the actual metadata payload,
        usually a Link or Layout object.
    signatures: A list of signatures over the canonical JSON representation
        of the value of the signed attribute.
    compact_json: A boolean indicating if the dump method should write a
        compact JSON string representation of the metadata.

  """
  signatures = attr.ib()
  signed = attr.ib()


  def __init__(self, **kwargs):
    self.signatures = kwargs.get("signatures", [])
    self.signed = kwargs.get("signed")
    self.compact_json = kwargs.get("compact_json", False)

    self.validate()


  def __repr__(self):
    """Returns the JSON string representation. """
    indent = None if self.compact_json else 1
    separators = (',', ':') if self.compact_json else (',', ': ')

    return json.dumps(
        {
          "signatures": self.signatures,
          "signed": attr.asdict(self.signed)
        },
        indent=indent,
        separators=separators,
        sort_keys=True
      )


  def dump(self, path):
    """Writes the JSON string representation of the instance to disk.

    Arguments:
      path: The path to write the file to.

    Raises:
      IOError: File cannot be written.

    """
    with open(path, "wb") as fp:
      fp.write("{}".format(self).encode("utf-8"))


  @staticmethod
  def load(path):
    """Loads the JSON string representation of in-toto metadata from disk.

    Arguments:
      path: The path to read the file from.

    Raises:
      IOError: The file cannot be read.
      securesystemslib.exceptions.FormatError: Metadata format is invalid.

    Returns:
      A Metablock object whose signable attribute is either a Link or a Layout
      object.

    """
    with open(path, "r", encoding="utf8") as fp:
      data = json.load(fp)

    signatures = data.get("signatures", [])
    signed_data = data.get("signed", {})
    signed_type = signed_data.get("_type")

    if signed_type == "link":
      signed = Link.read(signed_data)

    elif signed_type == "layout":
      signed = Layout.read(signed_data)

    else:
      raise securesystemslib.exceptions.FormatError("Invalid Metadata format")

    return Metablock(signatures=signatures, signed=signed)


  @property
  def type_(self):
    """A shortcut to the `type_` attribute of the object on the signable
    attribute (should be one of "link" or "layout"). """
    # NOTE: Trailing underscore is used by convention (pep8) to avoid conflict
    # with Python's type keyword.
    return self.signed.type_


  def sign(self, key):
    """Creates signature over signable with key and adds it to signatures.

    Uses the UTF-8 encoded canonical JSON byte representation of the signable
    attribute to create signatures deterministically.

    Attributes:
      key: A signing key. The format is securesystemslib.formats.KEY_SCHEMA.

    Raises:
      securesystemslib.exceptions.FormatError: Key argument is malformed.
      securesystemslib.exceptions.CryptoError, \
              securesystemslib.exceptions.UnsupportedAlgorithmError:
          Signing errors.

    Returns:
      The signature. Format is securesystemslib.formats.SIGNATURE_SCHEMA.

    """
    securesystemslib.formats.KEY_SCHEMA.check_match(key)

    signature = securesystemslib.keys.create_signature(key,
        self.signed.signable_bytes)

    self.signatures.append(signature)

    return signature

  def sign_gpg(self, gpg_keyid=None, gpg_home=None):
    """Creates signature over signable with gpg and adds it to signatures.

    Uses the UTF-8 encoded canonical JSON byte representation of the signable
    attribute to create signatures deterministically.

    Arguments:
      gpg_keyid (optional): A keyid used to identify a local gpg signing key.
          If omitted the default signing key is used.

      gpg_home (optional): A path to the gpg home directory. If not set the
          default gpg home directory is used.

    Raises:
      ValueError, OSError, securesystemslib.gpg.exceptions.CommandError, \
            securesystemslib.gpg.exceptions.KeyNotFoundError:
        gpg signing errors.

    Side Effects:
      Calls system gpg command in a subprocess.

    Returns:
      The signature. Format is securesystemslib.formats.GPG_SIGNATURE_SCHEMA.

    """
    signature = securesystemslib.gpg.functions.create_signature(
        self.signed.signable_bytes, gpg_keyid, gpg_home)

    self.signatures.append(signature)

    return signature


  def verify_signature(self, verification_key):
    """Verifies a signature over signable in signatures with verification_key.

    Uses the UTF-8 encoded canonical JSON byte representation of the signable
    attribute to verify the signature deterministically.

    NOTE: Only the first signature in the signatures attribute, whose keyid
    matches the verification_key keyid, is verified. If the verification_key
    format is securesystemslib.formats.GPG_PUBKEY_SCHEMA, subkey keyids are
    considered too.

    Arguments:
      verification_key: A verification key. The format is
          securesystemslib.formats.ANY_VERIFICATION_KEY_SCHEMA.

    Raises:
      securesystemslib.exceptions.FormatError: The passed key is malformed.

      SignatureVerificationError: No signature keyid matches the verification
          key keyid, or the matching signature is malformed, or the matching
          signature is invalid.

      securesystemslib.gpg.exceptions.KeyExpirationError: Passed verification
          key is an expired gpg key.

    """
    securesystemslib.formats.ANY_VERIFICATION_KEY_SCHEMA.check_match(
        verification_key)
    verification_keyid = verification_key["keyid"]

    # Find a signature that corresponds to the keyid of the passed
    # verification key or one of its subkeys
    signature = None
    for signature in self.signatures:
      if signature["keyid"] == verification_keyid:
        break

      if signature["keyid"] in list(
          verification_key.get("subkeys", {}).keys()):
        break

    else:
      raise SignatureVerificationError("No signature found for key '{}'"
          .format(verification_keyid))

    if securesystemslib.formats.GPG_SIGNATURE_SCHEMA.matches(signature):
      valid = securesystemslib.gpg.functions.verify_signature(signature,
          verification_key, self.signed.signable_bytes)

    elif securesystemslib.formats.SIGNATURE_SCHEMA.matches(signature):
      valid = securesystemslib.keys.verify_signature(
          verification_key, signature, self.signed.signable_bytes)

    else:
      valid = False

    if not valid:
      raise SignatureVerificationError("Invalid signature for keyid '{}'"
          .format(verification_keyid))


  def _validate_signed(self):
    """Private method to check if the 'signed' attribute contains a valid
    Layout or Link object. """

    if not (isinstance(self.signed, Layout) or isinstance(self.signed, Link)):
      raise securesystemslib.exceptions.FormatError("The Metblock's 'signed'"
        " property has has to be of type 'Link' or 'Layout'.")

    # If the signed object is a Link or Layout object validate it.
    self.signed.validate()


  def _validate_signatures(self):
    """Private method to check that the 'signatures' attribute is a list of
    signatures in the format 'securesystemslib.formats.ANY_SIGNATURE_SCHEMA'.
    """

    if not isinstance(self.signatures, list):
      raise securesystemslib.exceptions.FormatError("The Metablock's"
        " 'signatures' property has to be of type 'list'.")

    for signature in self.signatures:
      securesystemslib.formats.ANY_SIGNATURE_SCHEMA.check_match(signature)
